from sklearn import preprocessing
import numpy as np
from scipy.stats.mstats import rankdata as ranksample

#Normalization functions
def norm_minmax(x, min=0, max=1):
	#This function is useful for sparse data with lots of zeroes
	##print("....by minmax, min {0} and max {1}".format(min,max))
	min_max_scaler = preprocessing.MinMaxScaler(feature_range=(min,max))
	x_minmax = min_max_scaler.fit_transform(x) #Does it across each feature (i.e. column)
	return [x_minmax, min_max_scaler]

#def norm_scaler_reapply(scaler_func, x):
#        x_transformed = scaler_func.transform(x)
#        return x_transformed

def norm_rasterize(x_sample):
	#x_sample is an ndarray, of dimension (#features x 1)
	#Remember that newlist=list(oldlist) will still be mutable for nested lists in oldlist
	rasterize_range = range(1,len(x_sample)+1)
	ycopy=x_sample.copy(); prev=0
	for ordered_value in sorted(set(x_sample)):
		used_range = rasterize_range[prev:prev+len(x_sample[x_sample==ordered_value])]
		ycopy[x_sample==ordered_value] = sum(used_range)/len(used_range)
		prev = prev + len(x_sample[x_sample==ordered_value])
	#print(len(ycopy))
	return ycopy

def apply_norm_func(norm_func, xdat, bysamples=1):
        #If bysamples is 0, xdat[0] must be train_x, and xdat[1] must be test_x
	#If bysamples is 1, xdat is #samples x #features (not feat x samples, #Edit 03 Feb, 2017 | jgrewal)
	result = []
	if norm_func not in ["none", "rastminmax"]:
		raise ValueError('Incorrect normalization function, {0} passed'.format(norm_func))
	if norm_func=="none":
		result=xdat

	if norm_func=="rastminmax":
		temprast=[]
		#First rasterize
		if(type(xdat)==list):
			#xdat is sample x features, then for each sample
			#xdat is features x samples, then for each feature (bysamples==0)
			result=[]
			for xtype in xdat:
				temp = (map(norm_rasterize,xtype))
				temprast.append(temp)
		else:
			##print("Rasterized {0}".format(xdat.shape))
			temprast = ranksample(xdat, axis=bysamples) #(map(norm_rasterize,xdat))
			
		#Then minmax per sample (if bysamples==1) or by features (if bysamples==0)
		#if bysamples==0:
		#	[train_x_norm, scalefunc] = norm_minmax(x=temprast[0])
		#	test_x_norm=norm_scaler_reapply(scalefunc, temprast[1])
		#	valid_x_norm=norm_scaler_reapply(scalefunc, temprast[2])
		#	result=[train_x_norm, test_x_norm, valid_x_norm]
		#else:
		if bysamples==1:
			temprast = np.asarray(temprast)
			temprast = np.transpose(temprast)
			##print("Minmax'ing across each column {0}".format(temprast.shape))
			##print(temprast);
			##print(temprast.shape)
			x_norm = norm_minmax(x=temprast)[0]
			result = x_norm
			##print(result.shape)
			##Edit 03 Feb, 2017 | jgrewal
			result = result.transpose()
	
	if type(result) != list:
		print("norm result shape {0}".format(result.shape))
	return result





import lasagne
import theano
import theano.tensor as T
import numpy as np
from lasagne.utils import floatX
from lasagne.regularization import regularize_layer_params_weighted, l2, l1
from lasagne.regularization import regularize_layer_params

#Define a custom autoencoder (called during training)
def build_custom_autoencoder(n_encoded, num_features, target_var, learn_rate=0.001, input_var=None, is_image=False, keep_positive=False, l2_reg=0.001):
	if is_image:
		l_in = lasagne.layers.InputLayer(shape=(None, 1, num_features, 1), input_var=input_var)
	else:
		l_in = lasagne.layers.InputLayer(shape=(None, num_features), input_var=input_var)
	smallest_size = 2	
	nonlin_encode = lasagne.nonlinearities.sigmoid #linear
	nonlin_decode = lasagne.nonlinearities.rectify #sigmoid
	init_weights = lasagne.init.Constant(val=1.0) 
	init_weights = lasagne.init.Normal(std=1, mean = 1)
	#init_weights = lasagne.init.GlorotUniform()
	init_weights1 = theano.shared(floatX(init_weights((num_features, n_encoded))))
	init_weights2 = theano.shared(floatX(init_weights((n_encoded, n_encoded/2))))
	init_weights3 = theano.shared(floatX(init_weights((n_encoded/2, smallest_size))))
	print("Encode nonlin: {0}\nDecode nonlin: {1}\nInit weights type:{2}".format(nonlin_encode,nonlin_decode,init_weights))	
	if keep_positive:
		init_weights = theano.shared(floatX(init_weights((num_features, n_encoded))))
		init_weights = T.exp(init_weights)
		print(init_weights)	
	#Encoder step
	encoder_l_out1 = lasagne.layers.DenseLayer(l_in, num_units=n_encoded, W = init_weights1, nonlinearity=nonlin_encode, b=lasagne.init.Constant(0.))
	
	encoder_l_out2 = lasagne.layers.DenseLayer(encoder_l_out1, num_units=n_encoded/2, W = init_weights2, nonlinearity=nonlin_encode, b=lasagne.init.Constant(0.))
	encoder_l_out = lasagne.layers.DenseLayer(encoder_l_out2, num_units=smallest_size, W = init_weights3, nonlinearity=nonlin_encode, b=lasagne.init.Constant(0.))
		
	#Decoder step
	#	decoder_l_out = lasagne.layers.DenseLayer(encoder_l_out, num_units = num_features/2, W=encoder_l_out2.W.T, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))
	decoder_l_out = TransposedDenseLayer(encoder_l_out, num_units = n_encoded/2, W=encoder_l_out.W, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))
	decoder_l_out =  TransposedDenseLayer(decoder_l_out, num_units = n_encoded, W=encoder_l_out2.W, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))
	decoder_l_out =  TransposedDenseLayer(decoder_l_out, num_units = num_features, W=encoder_l_out1.W, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))	
	#	decoder_l_out = lasagne.layers.DenseLayer(dencoder_l_out, num_units = num_features, W=encoder_l_out1.W.T, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))
	
	##Decoder with tied weights
	#First add inverting dense layer
	#decoder_l_out = lasagne.layers.InverseLayer(encoder_l_out, encoder_l_out)
	###decoder_l_out = TransposedDenseLayer(encoder_l_out, num_units = num_features, W=encoder_l_out.W, nonlinearity=nonlin_decode, b=lasagne.init.Constant(0.))
	#decoder_l_out = lasagne.layers.DenseLayer(encoder_l_out, num_units = num_features, W=encoder_l_out.W.T, nonlinearity=nonlin_decode)
	
	#Define some theano vars
	target_values = target_var
	encoded_output = lasagne.layers.get_output(encoder_l_out)
	network_output = lasagne.layers.get_output(decoder_l_out)
	
	cost = lasagne.objectives.squared_error(network_output, target_values).mean() + regularize_layer_params(decoder_l_out, l2) * l2_reg
	all_params = lasagne.layers.get_all_params(decoder_l_out, trainable=True)
	
	#AdaDelta updates for training
	updates = lasagne.updates.rmsprop(cost, all_params, learning_rate=learn_rate)
	##updates = lasagne.updates.adadelta(cost, all_params)
	##updates = lasagne.updates.sgd(cost, all_params, learning_rate=learn_rate)
	
	#Theano functions for key queries
	train = theano.function([l_in.input_var, target_values], cost, updates=updates, allow_input_downcast=True)
	network = decoder_l_out
	predict = theano.function([l_in.input_var], network_output, allow_input_downcast=True)
	encode = theano.function([l_in.input_var], encoded_output, allow_input_downcast=True)
	return network, train, predict, encode

#Define a custom mlp (called during training)
def build_custom_mlp(n_out, num_features, depth, width, drop_input, drop_hidden, input_var=None, is_image=False):
	if is_image:
		network = lasagne.layers.InputLayer(shape=(None, 1, num_features,1), input_var=input_var)
	else:
		network = lasagne.layers.InputLayer(shape=(None,num_features), input_var=input_var)
	if drop_input:
		network = lasagne.layers.dropout(network, p=drop_input)
	nonlin = lasagne.nonlinearities.tanh
	for _ in range(depth):
		network=lasagne.layers.DenseLayer(network, width, nonlinearity=nonlin)
		if drop_hidden:
			network = lasagne.layers.dropout(network, p=drop_hidden)
	#Final output layer is softmax, for multiclass output
	softmax = lasagne.nonlinearities.softmax
	network = lasagne.layers.DenseLayer(network, n_out, nonlinearity=softmax)
	return network


#Define iterator over minibatches of inputs (x) and corresponding labels (y)
def iterate_minibatches(inputs, targets, batchsize, shuffle=False):
	assert len(inputs) == len(targets)
	if shuffle:
		indices = np.arange(len(inputs))
		np.random.shuffle(indices)
	for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
		if shuffle:
			excerpt = indices[start_idx:start_idx + batchsize]
        	else:
			excerpt = slice(start_idx, start_idx + batchsize)
	yield inputs[excerpt], targets[excerpt]

#Define function to reshape 1 dimensional input to an image with 1 channel and width 1
def make_1d_image(x):
	return(x.reshape((x.shape[0],1,x.shape[1],1)))

class TransposedDenseLayer(lasagne.layers.DenseLayer):
	def __init__(self, incoming, num_units, W=lasagne.init.GlorotUniform(), b=lasagne.init.Constant(0.), nonlinearity=lasagne.nonlinearities.rectify,**kwargs):
		super(TransposedDenseLayer, self).__init__(incoming, num_units, W, b, nonlinearity, **kwargs)
	
	def get_output_shape_for(self, input_shape):
		return (input_shape[0], self.num_units)
	
	def get_output_for(self, input, **kwargs):
		if input.ndim > 2:
			input = input.flatten(2)
		activation = T.dot(input, self.W.T)
		if self.b is not None:
			activation = activation + self.b.dimshuffle('x', 0)
		return self.nonlinearity(activation)



import os, sys
import glob, yaml, gzip
import numpy as np
import pandas as pd

###DEPRECATED lasagne + theano dependences
import lasagne
import theano
import theano.tensor as T
### Set theano flag to cpu to prevent cudnn issues with other users
os.environ['THEANO_FLAGS'] = "device=cpu"
#Change default theano GC for memory optimization
theano.config.allow_gc = True
theano.config.allow_pre_alloc = False
theano.config.scan.allow_gc = True
#END#OF#DEPRECATED lasagne + theano dependences

from .scope_io_modules import read_input, map_train_test_features_x, map_gene_names
from .scope_normalization_functions import *
from .scope_ensemble_functions import getmodelsdict, get_ensemble_score, build_custom_mlp
from .scope_plots import *
import cancerscope
from .config import SCOPEMODELS_DATADIR, SCOPEMODELS_LIST, SCOPEMODELS_FILELIST_DIR
import heapq
import copy
import gc

### Check if models have been downloaded already, and if not, do it!
if os.path.isdir(SCOPEMODELS_DATADIR) is False:
	"""If not already downloaded to pckg site, retrieve the models"""
	print("Thankyou for using cancerscope. The initial run requires download of dependent model files. Proceeding with download now...\n\tModels will be downloaded to {0}".format(SCOPEMODELS_DATADIR))
	cancerscope.downloadmodel(targetdir=SCOPEMODELS_DATADIR)
	### Collate the directories of all the models
	modeldirs_dict = cancerscope.getmodel()
	print("Models are downloaded at {0}".format(modeldirs_dict))

class scope(object):
	def __init__(self):
		self.models_dict = cancerscope.getmodelsdict()
		self.downloaded_models_dict = {}
		[self.downloaded_models_dict.update(m) for m in cancerscope.get_models.getmodel().values()]
		self.model_names = self.downloaded_models_dict.keys()
	def load_data(self, X_file):
		x_dat, x_samples, x_features, x_features_genecode = read_input(X_file)
		sys.stdout.write("\nRead in sample file {0}, \n\tData shape {1}\n\tNumber of samples {2}\n\tNumber of genes in input {3}, with gene code {4}".format(X_file, x_dat.shape, len(x_samples), len(x_features), x_features_genecode))
		return [x_dat, x_samples, x_features, x_features_genecode]
	def predict(self, X, x_features, x_features_genecode, x_sample_names=None, outdir=None, modelnames=None, get_preds_dict=False, return_plotdf=False):
		self.predict_dict = {}
		my_model_list = self.model_names
		if modelnames is not None:
			my_model_list = modelnames
		## Iterating over each model in the ensemble,
		for k_model in my_model_list:
			## Set up each individual model (clas scopemodel)
			lmodel = cancerscope.scopemodel(self.downloaded_models_dict[k_model])
			lmodel.fit()
			## Map training features to the genecode in the input
			mapped_model_features = map_gene_names(lmodel.features, genecode_out = x_features_genecode, genecode_in = "SCOPE")
			feat_subset_x = map_train_test_features_x(X, training_features = mapped_model_features, test_features = x_features, fillzero=True, defaultvalue=0.0)
			self.predict_dict[k_model] = lmodel.predict(feat_subset_x, get_predictions_dict=True)
			### GC to free up memory
			lmodel = None
			for i in range(0,3):
				gc.collect()
		if get_preds_dict is True:
			return self.predict_dict
		if return_plotdf is True:
			ens_df = cancerscope.scope_plots.get_plotting_df(self.predict_dict, x_sample_names = x_sample_names)
		else:
			## Do something to process output for ensemble score  
			ens_df = get_ensemble_score(self.predict_dict)
			if x_sample_names is not None:
				ens_df['sample_name'] = [x_sample_names[m] for m in ens_df['sample_ix'].tolist()]
			if outdir is not None:
				sys.stdout.write("\nOutdir provided, so writing prediction dataframe and plotting background data to files\n\tAlso generating per-sample plots at {0}".format(outdir))
				plot_bg_df = cancerscope.scope_plots.get_plotting_df(self.predict_dict, x_sample_names = x_sample_names)
				cancerscope.scope_plots.plot_cases(ens_df, plot_bg_df, outdir=outdir, save_txt=True)
		return ens_df
	
	def get_predictions_from_file(self, X_file, outdir=None, modelnames=None, get_preds_dict=False, return_plotdf=False):
		x_input, x_samples, x_features, x_features_genecode = self.load_data(X_file)
		prediction_df = self.predict(X=x_input, x_features = x_features, x_features_genecode = x_features_genecode, x_sample_names=x_samples, outdir=outdir, modelnames=modelnames, get_preds_dict=get_preds_dict, return_plotdf=return_plotdf)
		return(prediction_df)

class scopemodel(object):
	def __init__(self, modeldir):
		self.in_modeldir = modeldir
	def load_model(self,indir):
		in_model_npz = glob.glob(indir + "/lasagne_bestparams.npz")[0]
		in_labdict = glob.glob(indir + "/dataset_train/"+ "*/dict_labels*"+".txt")[0]
		in_trainidx = glob.glob(indir + "/dataset_train/" + "*/trainingidxs.txt")[0]
		in_trainfeatures = glob.glob(indir + "/featurelist.txt")[0] 
		myargs = self.get_params()
		with np.load(in_model_npz) as f: model_params = [f['arr_%d' % i] for i in range(len(f.files))]
		self.trainingdat = myargs['input'][0]
		with open(in_trainfeatures) as f: self.features = f.read().splitlines()
		n_in = model_params[0].shape[0]
		n_class = model_params[-1].shape[0]
		n_hiddenlayers=(len(model_params)/2)-1
		n_hidden = model_params[0].shape[1]
		self.normalization_input = myargs['normtype'][0]
		try:
			with(open(in_labdict)) as f:
				self.numtolabel = dict((int(v.rstrip()), k.rstrip()) for k,v in (line.split('\t') for line in f))
		except ValueError:
			with(open(in_labdict)) as f:
				self.numtolabel = dict((int(k.rstrip()), v.rstrip()) for k,v in (line.split('\t') for line in f))
		input_var= T.matrix('inputs')
		network = build_custom_mlp(input_var=input_var, is_image=False, n_out=n_class, num_features=n_in, depth=n_hiddenlayers, width=n_hidden, drop_hidden=0, drop_input=0)
		lasagne.layers.set_all_param_values(network, model_params)
		test_fn = theano.function([input_var], lasagne.layers.get_output(network, deterministic=True), allow_input_downcast=True)
		test_max_fn = theano.function([input_var], T.argmax(lasagne.layers.get_output(network, deterministic=True), axis=1), allow_input_downcast=True)
		return([network, test_fn, test_max_fn])	
	def prepare_input_featorders(self, X, x_features_genecode, x_features):
		#### Map training features to the genecode in the input
		mapped_model_features = map_gene_names(self.features, genecode_in = "SCOPE", genecode_out = x_features_genecode)
		feat_subset_x = map_train_test_features_x(X, mapped_model_features, x_features) ## This function will reorder the input based on the training features order, and if missing some genes, will set those to 0)
		return feat_subset_x
	def predict(self,X, get_all_predictions=True, get_numeric=True, get_predictions_dict=True):
		## First make sure X is filtered properly 
		if X.shape[1] != len(self.features):
			sys.stdout.write("\nReminder...Did you use the scopemodel.prepare_input_featorders(X, x_features_genecode, x_features) function before predict()?")
		X_normed = self.get_normalized_input(X)
		all_predicted = self.pred_fn(X_normed)
		max_predicted_classnum = self.pred_max_fn(X_normed)
		if get_predictions_dict is True:
			prediction_labs = [[self.numtolabel[i] for i,j in enumerate(m)] for m in all_predicted]
			sorted_list_of_lab_preds_sublists = [sorted(m, key=lambda x:x[1], reverse=True) for m in [zip(a,b) for a,b in zip(prediction_labs, all_predicted)]]
			#return [prediction_labs, all_predicted]
			return(np.asarray(sorted_list_of_lab_preds_sublists))
		else:
			if get_all_predictions is False:
				if get_numeric is False:
					return([self.numtolabel[i] for i in max_predicted_classnum])
				else:
					max_predicted_index_probability = np.amax(all_predicted, axis=1)
					return(max_predicted_index_probability)
			else:
				if get_numeric is False:
					return [[self.numtolabel[i] for i,j in enumerate(m)] for m in all_predicted]
				else:
					return(all_predicted)
	def fit(self):
		model_loaded = self.load_model(self.in_modeldir)
		self.network = model_loaded[0]
		self.pred_fn = model_loaded[1]
		self.pred_max_fn = model_loaded[2]
	def get_params(self):
		in_args = glob.glob(self.in_modeldir + "/argparse_args.txt")[0]
		with open(in_args) as f:
			myargs = yaml.safe_load(f) #March 17 2020 change from yaml.load(open(in_args), Loader=yaml.BaseLoader)
		return(myargs)
	def get_normalized_input(self,X):
		print("...Normalization function being applied: {0}".format(self.normalization_input))
		x_test = apply_norm_func(norm_func=self.normalization_input, bysamples=1, xdat=(X))
		return(x_test)
	def get_jacobian(self,X_test):
		X_normed = self.get_normalized_input(X_test)
		input_var = T.matrix('inputs')
		layers = lasagne.layers.get_all_layers(self.network)
		outputs = lasagne.layers.get_output(layers, input_var)
		output_final = T.flatten(outputs[-1])
		output_hidden = outputs[-2]
		output_initial = outputs[0]
		jacobian = theano.gradient.jacobian(output_final, output_initial)
		run = theano.function([input_var], jacobian, allow_input_downcast=True)
		outjac=run(X_normed)
		#As this tensor is padded with zero matrices on the non-diagnols, convert it to a stacked set of per_sample_per_output x input Jacobians
		outjac=remove_extra_submats(outjac)
		return(outjac)

def remove_extra_submats(bigmat):
	#As the theano jacobian calculation appends null matrices at the non-diagonals, this function returns a stacked set of (output_nodes * number samples) x (input_nodes) as a 2d matrix, from a 3d matrix returned by the jacobian gradient function in theano
	in_samples = bigmat.shape[1]
	out_nodes = int(bigmat.shape[0] / in_samples)
	print("BIGMAT SHAPE IS"); print(bigmat.shape)
	mymat = bigmat[0:out_nodes,0,:]
	for i in range(1, in_samples):
		mymat =np.append(mymat, bigmat[(0 + out_nodes*i):(out_nodes + out_nodes*i), i, :], axis=0)
	return(mymat)

 



import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import label_binarize
from itertools import cycle
from scipy import interp

import numpy as np
import os, glob, pandas, re
def plot_roc(pc, discrete_real_y, y_score, labdict, outdir=None,outname=None, cmap=plt.cm.Blues, title=None, plot_numeric=False):
	#discrete_real_y : Discrete valued class labels (min 0, max = #classes - 1). Dim = (#samples)
	#y_score : not singular prediction, but prediction score for each class for each sample, ordered 0 to #classes - 1. Dim = (#samples, #classes)
	#Binarize input real set
	n_classes = len(set(discrete_real_y))
	if(n_classes != y_score.shape[1]):
		print(n_classes)
		print(y_score.shape[1])
		print("ERROR: Input test range does not correspond to output scores range. Exiting ROC plot")
		return 0 
	
	real_y = label_binarize(discrete_real_y, classes=np.arange(n_classes))
	#Plotting defaults	
	plt.figure(pc, figsize=(20,20))
	np.set_printoptions(precision=2)
	#plt.tight_layout()
	plt.ylabel('TPR')
	plt.xlabel('FPR')
	plt.title(title + '\nReceiver operating characteristic curve')
	plt.xlim([0.0,1.0])
	plt.ylim([0.0,1.0])
	
	#Calculate the fpr and tprs for each label
	fpr=dict(); tpr=dict(); roc_auc=dict()
	for i in range(n_classes):
		fpr[i], tpr[i], _ = roc_curve(real_y[:, i], y_score[:, i])
		roc_auc[i] = auc(fpr[i], tpr[i])
	
	fpr["micro"], tpr["micro"], _ = roc_curve(real_y.ravel(), y_score.ravel())
	roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])
		
	#Aggregate all false positive rates
	all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
	#Get ROC curves for each class
	mean_tpr = np.zeros_like(all_fpr)
	for i in range(n_classes):
		mean_tpr += interp(all_fpr, fpr[i], tpr[i])
	#Calculate AUCs
	mean_tpr /= n_classes
	fpr["macro"] = all_fpr
	tpr["macro"] = mean_tpr
	roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])
	
	#Plot data
	plt.plot(fpr["micro"], tpr["micro"], label='micro averaged ROC curve, area {0:0.2f})'.format(roc_auc["micro"]), color='deeppink', linestyle=":", linewidth=4)
	plt.plot(fpr["macro"], tpr["macro"], label='macro averaged ROC curve, area {0:0.2f})'.format(roc_auc["macro"]), color='navy', linestyle=":", linewidth=4)	
	colors = cycle(['aqua', 'blue', 'cornflowerblue'])
	for i, color in zip(range(n_classes),colors):
		if roc_auc[i] < 0.51:
			plt.plot(fpr[i], tpr[i], color=color, lw=1, label='ROC curve of class {0}, area {1:0.2f}'.format(labdict[i], roc_auc[i]))
	plt.plot([0,1],[0,1], 'k--', lw=2)
	plt.legend(loc='lower right')
	
	if outdir is not None:
		plt.savefig(outdir+"/" + outname + ".png",facecolor='w', dpi=100)
	with open(outdir + "/" + outname  + "_eachclass.txt", 'w') as f_out:
		for key in range(n_classes):
			f_out.write('{0}\t{1}\n'.format(labdict[i], roc_auc[i]))
	
	with open(outdir + "/" + outname  + "_average.txt", 'w') as f_out:
		f_out.write('macro\t{0}'.format(roc_auc["macro"]))
		f_out.write('micro\t{0}'.format(roc_auc["micro"]))
		
	plt.close()

def plot_dir_confusion(indir, pattern=None, merge=False, plot_subs = False):
	#Plot confusion matrix after having generated the entire set of results
	#Finds a file called prediction_labels_*pattern*.txt, and fbeta_*pattern*.txt
	#And outputs a corresponding confusion matrix
	if merge :
		real = []; pred = []
		predfiles = [ indir+"/"+f for f in os.listdir(indir) if re.search(r'(prediction_labels_cv).*\.txt$', f)]
		fbetafile=glob.glob(indir+"/fbeta_full.txt")[0]
		for predfile in predfiles:
			dat = pandas.read_csv(predfile,sep="\t",header=None)
			real.append(dat[[2]][2].values.tolist())
			pred.append(dat[[0]][0].values.tolist())
		pattern = "merged_cv"; real=real[0]; pred=pred[0]
	else:
		if pattern is None:
			print("Input a pattern...")
			return 0	
		else:
			#Single file input
			predfile=glob.glob(indir+"/prediction_labels_"+pattern+".txt")[0]
			fbetafile=glob.glob(indir+"/fbeta_"+pattern+".txt")[0]
			dat = pandas.read_csv(predfile, sep="\t", header=None)
			real = dat[[2]][2].values.tolist()
			pred = dat[[0]][0].values.tolist()
	fscore = pandas.read_csv(fbetafile, sep="\t", header=None)[[0]]
	sorttarget=fscore[0].values.tolist()
	plot_confusion(pc=1, real_y=real, pred_y=pred, sorted_targetnames=sorttarget, outdir=indir, outname="confusion_"+pattern, title="", cmap=plt.cm.viridis)
	plt.close()
	if plot_subs :
		subset_truelabs = ["KICH", "HNSC_TS", "CESC", "ESCA", "THYM", "LUSC"]
		for matchname in subset_truelabs:
			indices = [i for i,x in enumerate(real) if re.search(matchname, x)]
			realx=[real[i] for i in indices]
			predx=[pred[i] for i in indices]
			plot_confusion(pc=1, real_y=realx, pred_y=predx, sorted_targetnames=sorttarget, outdir=indir, outname="confusion_"+pattern+"_" +matchname, title=matchname, cmap=plt.cm.viridis, plot_numeric=True)
			plt.close()	



import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix
import numpy as np
import os, glob, pandas, re
def plot_confusion(pc, real_y, pred_y, sorted_targetnames, outdir=None,outname=None, cmap=plt.cm.Blues, title=None, plot_numeric=False):
	plt.figure(pc, figsize=(20,20))
	alllabs = set(list(real_y) + list(pred_y))
	order_name = [x for x in sorted_targetnames if x in alllabs]
	
	cm_count=confusion_matrix(real_y, pred_y, labels=order_name)
	cm=cm_count.astype('float')/cm_count.sum(axis=1)[:,np.newaxis]
	
	np.set_printoptions(precision=2)
	plt.imshow(cm, interpolation='nearest', cmap=cmap)
	plt.title(title + "\nConfusion Matrix")
	plt.colorbar()
	tick_marks=np.arange(len(order_name))
	plt.xticks(tick_marks, order_name, rotation=90)
	plt.yticks(tick_marks, order_name)
	plt.tight_layout()
	plt.ylabel('True label')
	plt.xlabel('Predicted label')
	if plot_numeric:
		for i, cas in enumerate(cm_count):
			for j,c in enumerate(cas):
				if c>0:
					plt.text(j-.2, i+.2, float("{0:.3f}".format(c)), fontsize=18)
	if outdir is not None:
		plt.savefig(outdir+"/" + outname + ".png",facecolor='w', dpi=100)
	plt.close()

def plot_dir_confusion(indir, pattern=None, merge=False, plot_subs = False):
	#Plot confusion matrix after having generated the entire set of results
	#Finds a file called prediction_labels_*pattern*.txt, and fbeta_*pattern*.txt
	#And outputs a corresponding confusion matrix
	if merge :
		real = []; pred = []
		predfiles = [ indir+"/"+f for f in os.listdir(indir) if re.search(r'(prediction_labels_cv).*\.txt$', f)]
		fbetafile=glob.glob(indir+"/fbeta_full.txt")[0]
		for predfile in predfiles:
			dat = pandas.read_csv(predfile,sep="\t",header=None)
			real.append(dat[[2]][2].values.tolist())
			pred.append(dat[[0]][0].values.tolist())
		pattern = "merged_cv"; real=real[0]; pred=pred[0]
	else:
		if pattern is None:
			print("Input a pattern...")
			return 0	
		else:
			#Single file input
			predfile=glob.glob(indir+"/prediction_labels_"+pattern+".txt")[0]
			fbetafile=glob.glob(indir+"/fbeta_"+pattern+".txt")[0]
			dat = pandas.read_csv(predfile, sep="\t", header=None)
			real = dat[[2]][2].values.tolist()
			pred = dat[[0]][0].values.tolist()
	fscore = pandas.read_csv(fbetafile, sep="\t", header=None)[[0]]
	sorttarget=fscore[0].values.tolist()
	plot_confusion(pc=1, real_y=real, pred_y=pred, sorted_targetnames=sorttarget, outdir=indir, outname="confusion_"+pattern, title="", cmap=plt.cm.viridis)
	plt.close()
	if plot_subs :
		subset_truelabs = ["KICH", "HNSC_TS", "CESC", "ESCA", "THYM", "LUSC"]
		for matchname in subset_truelabs:
			indices = [i for i,x in enumerate(real) if re.search(matchname, x)]
			realx=[real[i] for i in indices]
			predx=[pred[i] for i in indices]
			plot_confusion(pc=1, real_y=realx, pred_y=predx, sorted_targetnames=sorttarget, outdir=indir, outname="confusion_"+pattern+"_" +matchname, title=matchname, cmap=plt.cm.viridis, plot_numeric=True)
			plt.close()	



import os
import glob
### Global variables 
SCOPEMODELS_DATADIR = os.path.abspath(os.path.dirname(__file__)) + "/data/"
SCOPEMODELS_FILELIST_DIR = os.path.abspath(os.path.dirname(__file__))
SCOPEMODELS_LIST = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'resources/scope_files.txt')
REF_LABELCOUNTS= "/".join([os.path.abspath(os.path.dirname(__file__)),'resources/dict_labcount_rm500.txt'])
REF_DISEASECODES= "/".join([os.path.abspath(os.path.dirname(__file__)),'resources/diseasetypes_codes.txt'])

### Compile list of models used in training  
def getmodelsdict():
	modelOptions = {}
	with open(SCOPEMODELS_LIST, 'r') as f:
		for line in f:
			if line.strip()!='':
				modelname, url, expectedFile, expectedmd5 = line.strip().split('\t')
				modelOptions[modelname] = (url, expectedFile, expectedmd5)
	return modelOptions




"""
This tutorial introduces logistic regression using Theano and stochastic
gradient descent.

Logistic regression is a probabilistic, linear classifier. It is parametrized
by a weight matrix :math:`W` and a bias vector :math:`b`. Classification is
done by projecting data points onto a set of hyperplanes, the distance to
which is used to determine a class membership probability.

Mathematically, this can be written as:

.. math::
  P(Y=i|x, W,b) &= softmax_i(W x + b) \\
                &= \frac {e^{W_i x + b_i}} {\sum_j e^{W_j x + b_j}}


The output of the model or prediction is then done by taking the argmax of
the vector whose i'th element is P(Y=i|x).

.. math::

  y_{pred} = argmax_i P(Y=i|x,W,b)


This tutorial presents a stochastic gradient descent optimization method
suitable for large datasets.


References:

    - textbooks: "Pattern Recognition and Machine Learning" -
                 Christopher M. Bishop, section 4.3.2

"""

from __future__ import print_function

__docformat__ = 'restructedtext en'

import six.moves.cPickle as pickle
import gzip
import os
import sys
import timeit

import numpy

import theano
import theano.tensor as T


class LogisticRegression(object):
    """Multi-class Logistic Regression Class

    The logistic regression is fully described by a weight matrix :math:`W`
    and bias vector :math:`b`. Classification is done by projecting data
    points onto a set of hyperplanes, the distance to which is used to
    determine a class membership probability.
    """

    def __init__(self, input, n_in, n_out, W=None, b=None):
        """ Initialize the parameters of the logistic regression

        :type input: theano.tensor.TensorType
        :param input: symbolic variable that describes the input of the
                      architecture (one minibatch)

        :type n_in: int
        :param n_in: number of input units, the dimension of the space in
                     which the datapoints lie

        :type n_out: int
        :param n_out: number of output units, the dimension of the space in
                      which the labels lie

        """
        # start-snippet-1
        # initialize with 0 the weights W as a matrix of shape (n_in, n_out)
	if W is None:
	        self.W = theano.shared(
			value=numpy.zeros(
				(n_in, n_out),
				dtype=theano.config.floatX
			),
			name='W',
			borrow=True
		)
	else:
		self.W = theano.shared(value=W, name='W', borrow=True)
	
        # initialize the biases b as a vector of n_out 0s
	if b is None:
		self.b = theano.shared(
			value=numpy.zeros(
				(n_out,),
				dtype=theano.config.floatX
			),
			name='b',
			borrow=True
		)
	else:
		self.b = theano.shared(value=b, name='b', borrow=True)

	
        # symbolic expression for computing the matrix of class-membership
        # probabilities
        # Where:
        # W is a matrix where column-k represent the separation hyperplane for
        # class-k
        # x is a matrix where row-j  represents input training sample-j
        # b is a vector where element-k represent the free parameter of
        # hyperplane-k
        self.p_y_given_x = T.nnet.softmax(T.dot(input, self.W) + self.b)

        # symbolic description of how to compute prediction as class whose
        # probability is maximal
        self.y_pred = T.argmax(self.p_y_given_x, axis=1)
        self.y_pred_score = T.max(self.p_y_given_x, axis=1)
        # end-snippet-1

        # parameters of the model
        self.params = [self.W, self.b]

        # keep track of model input
        self.input = input

    def negative_log_likelihood(self, y, misClassCost):
        """Return the mean of the negative log-likelihood of the prediction
        of this model under a given target distribution. We will also be adjusting the error based on the size of the training class (smaller class, magnify error)
        .. math::

            \frac{1}{|\mathcal{D}|} \mathcal{L} (\theta=\{W,b\}, \mathcal{D}) =
            \frac{1}{|\mathcal{D}|} \sum_{i=0}^{|\mathcal{D}|}
                \log(P(Y=y^{(i)}|x^{(i)}, W,b)) \\
            \ell (\theta=\{W,b\}, \mathcal{D})

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label

        Note: we use the mean instead of the sum so that
              the learning rate is less dependent on the batch size
        """
        # start-snippet-2
        # y.shape[0] is (symbolically) the number of rows in y, i.e.,
        # number of examples (call it n) in the minibatch
        # T.arange(y.shape[0]) is a symbolic vector which will contain
        # [0,1,2,... n-1] T.log(self.p_y_given_x) is a matrix of
        # Log-Probabilities (call it LP) with one row per example and
        # one column per class LP[T.arange(y.shape[0]),y] is a vector
        # v containing [LP[0,y[0]], LP[1,y[1]], LP[2,y[2]], ...,
        # LP[n-1,y[n-1]]] and T.mean(LP[T.arange(y.shape[0]),y]) is
        # the mean (across minibatch examples) of the elements in v,
        # i.e., the mean log-likelihood across the minibatch.
        return -T.mean(T.log(self.p_y_given_x)[T.arange(y.shape[0]), y] + T.mul(T.log(self.p_y_given_x)[T.arange(y.shape[0]), y], misClassCost))
        # end-snippet-2

    def errors(self, y):
        """Return a float representing the number of errors in the minibatch
        over the total number of examples of the minibatch ; zero one
        loss over the size of the minibatch

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label
        """

        # check if y has same dimension of y_pred
        if y.ndim != self.y_pred.ndim:
            raise TypeError(
                'y should have the same shape as self.y_pred',
                ('y', y.type, 'y_pred', self.y_pred.type)
            )
        # check if y is of the correct datatype
        if y.dtype.startswith('int'):
            # the T.neq operator returns a vector of 0s and 1s, where 1
            # represents a mistake in prediction
            return T.mean(T.neq(self.y_pred, y))
        else:
            raise NotImplementedError()
 

    ##Add weight saving code
    def getstate(self):
        state=[self.W.get_value(), self.b.get_value()]
        return state

    def setstate(self, state):
        W,b=state
        self.W = W
        self.b = b
        self.W = theano.shared(value=W, name='W', borrow=True)
        self.b = theano.shared(value=b, name='b', borrow=True)

def load_data(dataset, sharedflag=None):
    ''' Loads the dataset

    :type dataset: string
    :param dataset: the path to the dataset (here MNIST)
    '''

    #############
    # LOAD DATA #
    #############

    # Download the MNIST dataset if it is not present
    data_dir, data_file = os.path.split(dataset)
    if data_dir == "" and not os.path.isfile(dataset):
        # Check if dataset is in the data directory.
        new_path = os.path.join(
            os.path.split(__file__)[0],
            "..",
            "data",
            dataset
        )
        if os.path.isfile(new_path) or data_file == 'mnist.pkl.gz':
            dataset = new_path

    if (not os.path.isfile(dataset)) and data_file == 'mnist.pkl.gz':
        from six.moves import urllib
        origin = (
            'http://www.iro.umontreal.ca/~lisa/deep/data/mnist/mnist.pkl.gz'
        )
        print('Downloading data from %s' % origin)
        urllib.request.urlretrieve(origin, dataset)

    print('... loading data')

    # Load the dataset
    with gzip.open(dataset, 'rb') as f:
        try:
            train_set, valid_set, test_set = pickle.load(f, encoding='latin1')
        except:
            train_set, valid_set, test_set = pickle.load(f)
    # train_set, valid_set, test_set format: tuple(input, target)
    # input is a numpy.ndarray of 2 dimensions (a matrix)
    # where each row corresponds to an example. target is a
    # numpy.ndarray of 1 dimension (vector) that has the same length as
    # the number of rows in the input. It should give the target
    # to the example with the same index in the input.

    def shared_dataset(data_xy, borrow=True):
        """ Function that loads the dataset into shared variables

        The reason we store our dataset in shared variables is to allow
        Theano to copy it into the GPU memory (when code is run on GPU).
        Since copying data into the GPU is slow, copying a minibatch everytime
        is needed (the default behaviour if the data is not in a shared
        variable) would lead to a large decrease in performance.
        """
        data_x, data_y = data_xy
        shared_x = theano.shared(numpy.asarray(data_x,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        shared_y = theano.shared(numpy.asarray(data_y,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        # When storing data on the GPU it has to be stored as floats
        # therefore we will store the labels as ``floatX`` as well
        # (``shared_y`` does exactly that). But during our computations
        # we need them as ints (we use labels as index, and if they are
        # floats it doesn't make sense) therefore instead of returning
        # ``shared_y`` we will have to cast it to int. This little hack
        # lets ous get around this issue
        return shared_x, T.cast(shared_y, 'int32')

    if sharedflag is None:
        test_set_x, test_set_y = shared_dataset(test_set)
        valid_set_x, valid_set_y = shared_dataset(valid_set)
        train_set_x, train_set_y = shared_dataset(train_set)

        rval = [(train_set_x, train_set_y), (valid_set_x, valid_set_y),
            (test_set_x, test_set_y)]
    else:
        rval = [train_set, valid_set, test_set]
    return rval


def sgd_optimization_mnist(learning_rate=0.13, n_epochs=1000,
                           dataset='mnist.pkl.gz',
                           batch_size=600):
    """
    Demonstrate stochastic gradient descent optimization of a log-linear
    model

    This is demonstrated on MNIST.

    :type learning_rate: float
    :param learning_rate: learning rate used (factor for the stochastic
                          gradient)

    :type n_epochs: int
    :param n_epochs: maximal number of epochs to run the optimizer

    :type dataset: string
    :param dataset: the path of the MNIST dataset file from
                 http://www.iro.umontreal.ca/~lisa/deep/data/mnist/mnist.pkl.gz

    """
    datasets = load_data(dataset)

    train_set_x, train_set_y = datasets[0]
    valid_set_x, valid_set_y = datasets[1]
    test_set_x, test_set_y = datasets[2]

    # compute number of minibatches for training, validation and testing
    n_train_batches = train_set_x.get_value(borrow=True).shape[0] // batch_size
    n_valid_batches = valid_set_x.get_value(borrow=True).shape[0] // batch_size
    n_test_batches = test_set_x.get_value(borrow=True).shape[0] // batch_size

    ######################
    # BUILD ACTUAL MODEL #
    ######################
    print('... building the model')

    # allocate symbolic variables for the data
    index = T.lscalar()  # index to a [mini]batch

    # generate symbolic variables for input (x and y represent a
    # minibatch)
    x = T.matrix('x')  # data, presented as rasterized images
    y = T.ivector('y')  # labels, presented as 1D vector of [int] labels

    # construct the logistic regression class
    # Each MNIST image has size 28*28
    classifier = LogisticRegression(input=x, n_in=28 * 28, n_out=10)

    # the cost we minimize during training is the negative log likelihood of
    # the model in symbolic format
    cost = classifier.negative_log_likelihood(y)

    # compiling a Theano function that computes the mistakes that are made by
    # the model on a minibatch
    test_model = theano.function(
        inputs=[index],
        outputs=classifier.errors(y),
        givens={
            x: test_set_x[index * batch_size: (index + 1) * batch_size],
            y: test_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )

    validate_model = theano.function(
        inputs=[index],
        outputs=classifier.errors(y),
        givens={
            x: valid_set_x[index * batch_size: (index + 1) * batch_size],
            y: valid_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )

    # compute the gradient of cost with respect to theta = (W,b)
    g_W = T.grad(cost=cost, wrt=classifier.W)
    g_b = T.grad(cost=cost, wrt=classifier.b)

    # start-snippet-3
    # specify how to update the parameters of the model as a list of
    # (variable, update expression) pairs.
    updates = [(classifier.W, classifier.W - learning_rate * g_W),
               (classifier.b, classifier.b - learning_rate * g_b)]

    # compiling a Theano function `train_model` that returns the cost, but in
    # the same time updates the parameter of the model based on the rules
    # defined in `updates`
    train_model = theano.function(
        inputs=[index],
        outputs=cost,
        updates=updates,
        givens={
            x: train_set_x[index * batch_size: (index + 1) * batch_size],
            y: train_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )
    # end-snippet-3

    ###############
    # TRAIN MODEL #
    ###############
    print('... training the model')
    # early-stopping parameters
    patience = 5000  # look as this many examples regardless
    patience_increase = 2  # wait this much longer when a new best is
                                  # found
    improvement_threshold = 0.995  # a relative improvement of this much is
                                  # considered significant
    validation_frequency = min(n_train_batches, patience // 2)
                                  # go through this many
                                  # minibatche before checking the network
                                  # on the validation set; in this case we
                                  # check every epoch

    best_validation_loss = numpy.inf
    test_score = 0.
    start_time = timeit.default_timer()

    done_looping = False
    epoch = 0
    while (epoch < n_epochs) and (not done_looping):
        epoch = epoch + 1
        for minibatch_index in range(n_train_batches):

            minibatch_avg_cost = train_model(minibatch_index)
            # iteration number
            iter = (epoch - 1) * n_train_batches + minibatch_index

            if (iter + 1) % validation_frequency == 0:
                # compute zero-one loss on validation set
                validation_losses = [validate_model(i)
                                     for i in range(n_valid_batches)]
                this_validation_loss = numpy.mean(validation_losses)

                print(
                    'epoch %i, minibatch %i/%i, validation error %f %%' %
                    (
                        epoch,
                        minibatch_index + 1,
                        n_train_batches,
                        this_validation_loss * 100.
                    )
                )

                # if we got the best validation score until now
                if this_validation_loss < best_validation_loss:
                    #improve patience if loss improvement is good enough
                    if this_validation_loss < best_validation_loss *  \
                       improvement_threshold:
                        patience = max(patience, iter * patience_increase)

                    best_validation_loss = this_validation_loss
                    # test it on the test set

                    test_losses = [test_model(i)
                                   for i in range(n_test_batches)]
                    test_score = numpy.mean(test_losses)

                    print(
                        (
                            '     epoch %i, minibatch %i/%i, test error of'
                            ' best model %f %%'
                        ) %
                        (
                            epoch,
                            minibatch_index + 1,
                            n_train_batches,
                            test_score * 100.
                        )
                    )

                    # save the best model
                    with open('best_model.pkl', 'wb') as f:
                        pickle.dump(classifier, f)

            if patience <= iter:
                done_looping = True
                break

    end_time = timeit.default_timer()
    print(
        (
            'Optimization complete with best validation score of %f %%,'
            'with test performance %f %%'
        )
        % (best_validation_loss * 100., test_score * 100.)
    )
    print('The code run for %d epochs, with %f epochs/sec' % (
        epoch, 1. * epoch / (end_time - start_time)))
    print(('The code for file ' +
           os.path.split(__file__)[1] +
           ' ran for %.1fs' % ((end_time - start_time))), file=sys.stderr)


def predict():
    """
    An example of how to load a trained model and use it
    to predict labels.
    """

    # load the saved model
    classifier = pickle.load(open('best_model.pkl'))

    # compile a predictor function
    predict_model = theano.function(
        inputs=[classifier.input],
        outputs=classifier.y_pred)

    # We can test it on some examples from test test
    dataset='mnist.pkl.gz'
    datasets = load_data(dataset)
    test_set_x, test_set_y = datasets[2]
    test_set_x = test_set_x.get_value()

    predicted_values = predict_model(test_set_x[:10])
    print("Predicted values for the first 10 examples in test set:")
    print(predicted_values)


if __name__ == '__main__':
    sgd_optimization_mnist()



import lasagne
import theano
import theano.tensor as T
import numpy as np
from lasagne.utils import floatX
from lasagne.regularization import regularize_layer_params_weighted, l2, l1
from lasagne.regularization import regularize_layer_params

#Define a custom mlp (called during training)
def build_custom_mlp(n_out, num_features, depth, width, drop_input, drop_hidden, input_var=None, is_image=False):
	network={}
	if is_image:
		network['input'] = lasagne.layers.InputLayer(shape=(None, 1, num_features,1), input_var=input_var)
	else:
		network['input'] = lasagne.layers.InputLayer(shape=(None,num_features), input_var=input_var)
	if drop_input:
		network['input'] = lasagne.layers.dropout(network['input'], p=drop_input)
	
	nonlin = lasagne.nonlinearities.tanh
	for i in range(depth):
		if i == 0:
			network["hidden0"] = lasagne.layers.DenseLayer(network['input'],width, nonlinearity=nonlin) 
		else:
			network["hidden"+str(i)]=lasagne.layers.DenseLayer(network, width, nonlinearity=nonlin)
		
		if drop_hidden:
			network["hidden"+str(i)] = lasagne.layers.dropout(network["hidden"+str(i)], p=drop_hidden)
	
	#Final output layer is softmax, for multiclass output
	softmax = lasagne.nonlinearities.softmax
	network['output'] = lasagne.layers.DenseLayer(network["hidden"+str(i)], n_out, nonlinearity=softmax)
	return(network['output'], network)

def load_network(network_params_npz):
	with np.load(network_params_npz) as f:
		network_params = [f['arr_%d' % i] for i in range(len(f.files))]
	n_in = network_params[0].shape[0]; n_class = network_params[-1].shape[0]
	num_hidden_layers = (len(network_params)/2) - 1
	num_hidden_nodes = network_params[0].shape[1]
	target_var = T.ivector('targets'); input_var=T.matrix('inputs')
	ntwk = build_custom_mlp(input_var=input_var, n_out=n_class, num_features=n_in, depth=num_hidden_layers, width=num_hidden_nodes, drop_input=0.0, drop_hidden=0.0)
	return(ntwk)

def build_transfer_mlp(old_network, drop_input, drop_hidden, input_var=None, is_image=False):
	if os.path.isfile(old_network):
		old_network = load_network(old_network)
	

#Define iterator over minibatches of inputs (x) and corresponding labels (y)
def iterate_minibatches(inputs, targets, batchsize, shuffle=False):
	assert len(inputs) == len(targets)
	if shuffle:
		indices = np.arange(len(inputs))
		np.random.shuffle(indices)
	for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
		if shuffle:
			excerpt = indices[start_idx:start_idx + batchsize]
        	else:
			excerpt = slice(start_idx, start_idx + batchsize)
	yield inputs[excerpt], targets[excerpt]

#Define function to reshape 1 dimensional input to an image with 1 channel and width 1
def make_1d_image(x):
	return(x.reshape((x.shape[0],1,x.shape[1],1)))

class TransposedDenseLayer(lasagne.layers.DenseLayer):
	def __init__(self, incoming, num_units, W=lasagne.init.GlorotUniform(), b=lasagne.init.Constant(0.), nonlinearity=lasagne.nonlinearities.rectify,**kwargs):
		super(TransposedDenseLayer, self).__init__(incoming, num_units, W, b, nonlinearity, **kwargs)
	
	def get_output_shape_for(self, input_shape):
		return (input_shape[0], self.num_units)
	
	def get_output_for(self, input, **kwargs):
		if input.ndim > 2:
			input = input.flatten(2)
		activation = T.dot(input, self.W.T)
		if self.b is not None:
			activation = activation + self.b.dimshuffle('x', 0)
		return self.nonlinearity(activation)



import os, gzip, pickle
def load_custom_data(dataset):
        data_dir, data_file = os.path.split(dataset)
        if data_dir == "" and not os.path.isfile(dataset):
                new_path = os.path.join(
                 os.path.split(__file__)[0],
                "..",
                "data",
                dataset
                )
                if os.path.isfile(new_path):
                        dataset = new_path
         
        with gzip.open(dataset, 'rb') as f:
                try:
                        train_set, valid_set, test_set = pickle.load(f, encoding='latin1')
                except:
                        train_set, valid_set, test_set = pickle.load(f)
        
        return [train_set, valid_set, test_set]





#Contains commonly used classes and functions
#For I/O purposes in ML model runs

#Shared libs
import os, sys
import timeit
import pkgutil
import pandas as pd
import copy
import numpy as np
import operator

def map_train_test_features_x(x_in, training_features, test_features, fillzero=True, defaultvalue=0.0):
	x = copy.deepcopy(x_in)## Create another persistent copy of x as we are modifying it
	# Check that the matrix is ordered as samples x genes   
	assert(x.shape[1] == len(test_features))
	
	# For each training feature, get the corresponding index in the test features 
	reorder_test = {test_features.index(i):training_features.index(i) for i in test_features if i in training_features}
	### Get a list of indices indicating which index of the test features list corresponds to the corresponding ith' position in the training features list
	test_keys_orderedix = [i[0] for i in sorted(reorder_test.items(), key=operator.itemgetter(1))]
	
	num_testfeats_mapping_to_trainingfeats = len(test_keys_orderedix)
	#assert(test_features[test_keys_orderedix[0]] == training_features[0])
	
	testX_values = np.apply_along_axis(lambda x: x[test_keys_orderedix], 1,x)
	if num_testfeats_mapping_to_trainingfeats != len(training_features):
		sys.stdout.write("\n...Only {0} of input features mapped to expected number of features. Setting the rest to {1}".format(num_testfeats_mapping_to_trainingfeats, defaultvalue))
	if fillzero:
		for i in range(0, len(training_features), 1):
			if training_features[i] not in test_features:
				testX_values = np.insert(testX_values, i, defaultvalue, axis=1)
	return(testX_values)

def map_gene_names(features_list, genecode_in, genecode_out):
	### Options for genecode_in and genecode_out are ENSG, HUGO, GENCODE, GAF (and additional options from the resources file below)
	genedict = pd.read_csv("/".join([os.path.dirname(sys.modules["cancerscope"].__file__), 'resources/scope_features_genenames.txt']), delimiter="\t")
	if genecode_in == genecode_out:
		sys.stdout.write("\nWARNING! You passed the same genecode for input and output, does not need mapping")
		return features_list
	try:
		genecode_in in genedict.columns.tolist()
		genecode_out in genedict.columns.tolist()
	except (TypeError, NameError,ValueError):
		sys.stdout.write("Error: {1}\n".format(sys.exc_info()[0]))
		sys.stdout.write("Genecode should be one of {0}\n".format(genedict.columns.tolist()))
	else:
		genedict = genedict.set_index(genecode_in) 
		features_mapped = list(genedict.loc[features_list][genecode_out]) 
	try:
		len(set(features_mapped)) == len(features_list)
	except:
		sys.stdout.write("\nWARNING! The mapping could not be completed accurately\n")
	else:
		return features_mapped
	
def read_input(input_file, sep="\t"): 
	"""This function reads in the set of samples to predict. By default, the input is arranged as gene_name \t Sample 1 RPKM \t Sample 2 RPKM ....
	An alternative column separator can be passed to the function using the 'sep' attribute"""
	try:
		os.path.isfile(input_file)
	except:
		sys.stdout.write("\nError: {1}\n".format(sys.exc_info()[0]))
	else:
		sys.stdout.write("\nReading file {0}\n".format(input_file))
		genedict = pd.read_csv("/".join([os.path.dirname(sys.modules["cancerscope"].__file__), 'resources/scope_features_genenames.txt']), delimiter=sep)
		input_dat = pd.read_csv(input_file, delimiter=sep)
		in_genecode = input_dat.columns.tolist()[0] # What is the gene naming convention for the input  - this is the 1st column
	"""Feature subsetting"""
	try:
		in_genecode in ["GENCODE", genedict.columns.tolist()]
	except (TypeError, NameError,ValueError):
		sys.stdout.write("Error: {1}\n".format(sys.exc_info()[0]))
		sys.stdout.write("First column's name should be one of {0}\n".format(genedict.columns.tolist()))
	else:
		"""Separate out the input data into features (col 1), X (numeric values), and samples"""
		features_test = input_dat[in_genecode].tolist()
		X_data = input_dat.loc[:, input_dat.columns != in_genecode]
		samples = X_data.columns.tolist()
		X = X_data.values #to_records(index=False)
		if in_genecode == "GENCODE":
			sys.stdout.write("\nYour genenames are of format ENSG.version. Truncating to ENSG ID only\n")
			features_test = [m.split(".")[0] for m in features_test]
			in_genecode = "ENSEMBL"
		#sys.stdout.write("\nX shapei s {0} and features len is {1}".format(X.shape, len(features_test)))
	
	if(len(samples) == 1):
		## if there is only one sample, numpy array doesnt act nice, has to be reshaped
		X = X.reshape(1, len(X))
	else:
		## Otherwise, reshape as samples x genes, instead of genes x samples
		X = X.transpose()
	return X, samples, features_test, in_genecode




import copy
import pandas as pd
import numpy as np
import lasagne
import functools
#from numpy import divmod as numpydivmod

### List of models
def getmodelsdict():
        modelOptions = {}
        with open(SCOPEMODELS_LIST, 'r') as f:
                for line in f:
                        if line.strip()!='':
                                modelname, url, expectedFile, expectedmd5 = line.strip().split('\t')
                                modelOptions[modelname] = (url, expectedFile, expectedmd5)
        return modelOptions


### Function that enables removal of specific labels, or to sort the prediction dict by numeric (label prediction) or alphabet (labels)
def drop_sort_class(matrix4d_of_preds, label_to_drop=None, sortorder="alphabets"):
        ### This function takes in a matrix form of the dictionary containing individual models (shape num_models, num_samples, num_features, feature_tuples)
        ### It will return either a matrix form of the dictionary with the specified labels to drop removed, 
        ### Or optionally can return a amtrix form of the dictionary where the tuples are ordered by either 'numbers' or 'alphabets'
        if sortorder == 'numbers':
                ix_in_tuple_for_order = 1
        else:
                ix_in_tuple_for_order = 0
        df_matrix = copy.deepcopy(matrix4d_of_preds)
        if label_to_drop is None:
                for i in range(0, df_matrix.shape[0]):
                        for j in range(0, df_matrix.shape[1]):
                                sample_list_of_lab_tuples = df_matrix[i][j]
                                new_alphabet_order = sorted(sample_list_of_lab_tuples.tolist(), key=lambda x:-float(x[ix_in_tuple_for_order]))
                                df_matrix[i][j] = new_alphabet_order
                return df_matrix
        else:
                if ix_in_tuple_for_order == 1:
                        # We need to order the matrix alphabetically prior to deleting any labels
                        df_matrix = drop_sort_class(matrix4d_of_preds, label_to_drop=None, sortorder='alphabets')
                for p in label_to_drop:
                        ## DO something to drop the column
                        dropsite_labelix = np.where(df_matrix[0,0,:,0]==p)[0].tolist()
                        df_matrix = np.delete(df_matrix, dropsite_labelix, axis=2)
                return drop_sort_class(df_matrix, label_to_drop=None, sortorder=sortorder)

### Function that takes in dict and returns the top voted classes per sample and corresponding predictions
def get_ensemble_score(dict_with_preds, ignorelabs_list=None, x_sample_names=None):
        ### This function assumes the label,value tuples for each sample, for each model, are ordered by highest value (ix 0) to lowest value (ix -1) 
        modelnames = list(dict_with_preds.keys())
        df_temp_matrix = np.array([dict_with_preds[i] for i in modelnames])
        df_matrix = drop_sort_class(df_temp_matrix, label_to_drop=ignorelabs_list, sortorder='numbers')
        ### First get a list of predicted labels and values
        num_models, num_samples, num_preds, num_tuples_preds = df_matrix.shape
        flat_top_preds_values = df_matrix[:, :,0, 1].transpose().flatten() # Across all models, across all samples, the 0th ordered (top level) prediction, numeric value
        flat_top_preds_labels = df_matrix[:, :,0, 0].transpose().flatten() # Across all models, across all samples, the 0th ordered (top level) prediction, label
        ## Combine these into a sensible dataframe so as to separate each sample
        topPreds_bymodels_df = pd.DataFrame(np.column_stack([flat_top_preds_labels, flat_top_preds_values, modelnames * num_samples]), columns=['label', 'pred', 'modelname'])
        topPreds_bymodels_df['sample_ix'] = [divmod(m, num_models)[0] for m in topPreds_bymodels_df.index.tolist()]
        topPreds_bymodels_df[['pred']] = topPreds_bymodels_df[['pred']].astype(float) # dtype conversion for confidence scores
        ## Aggregate based on the predicted labels
        avg_per_label = topPreds_bymodels_df.groupby(['sample_ix', 'label'], as_index=False).mean()
        modelnames_label = topPreds_bymodels_df.groupby(['sample_ix', 'label'], as_index=False)['modelname'].apply(lambda x: "%s" % ','.join(x)).reset_index().rename(columns={0:"models"})
        modelnames_count = topPreds_bymodels_df.groupby(['sample_ix', 'label'], as_index=False)['modelname'].count().rename(columns={"modelname":"freq"})
        joined_df_list = [avg_per_label, modelnames_count ,modelnames_label]
        df_merged = functools.reduce(lambda  left,right: pd.merge(left,right,on=['sample_ix','label'], how='outer'), joined_df_list)
        df_merged.sort_values(by=['sample_ix', 'freq', 'pred'], ascending=False).reset_index(drop=True)        
        df_merged = df_merged.sort_values(by=['sample_ix', 'freq', 'pred'], ascending=False).reset_index(drop=True)
        df_merged["rank_pred"] = df_merged.index
        for s in range(0, num_samples):
                df_merged.loc[df_merged.sample_ix == s, "rank_pred"] =  range(1, df_merged[df_merged.sample_ix == s].shape[0]  +1)
        df_merged = df_merged.sort_values(by=['sample_ix', 'rank_pred'], ascending=True).reset_index(drop=True)
        if x_sample_names is not None:
                df_merged['sample_name'] = [x_sample_names[m] for m in df_merged['sample_ix'].tolist()]
        return df_merged

### Function to set up a custom network
def build_custom_mlp(n_out, num_features, depth, width, drop_input, drop_hidden, input_var=None, is_image=False):
        n_out = int(n_out); num_features=int(num_features); depth=int(depth); width=int(width); 
        if is_image:
                network = lasagne.layers.InputLayer(shape=(None, 1, num_features,1), input_var=input_var)
        else:
                network = lasagne.layers.InputLayer(shape=(None,num_features), input_var=input_var)
        if drop_input:
                network = lasagne.layers.dropout(network, p=drop_input)
        nonlin = lasagne.nonlinearities.tanh
        for _ in range(depth):
                network=lasagne.layers.DenseLayer(network, width, nonlinearity=nonlin)
                if drop_hidden:
                        network = lasagne.layers.dropout(network, p=drop_hidden)
        #Final output layer is softmax, for multiclass output
        softmax = lasagne.nonlinearities.softmax
        network = lasagne.layers.DenseLayer(network, n_out, nonlinearity=softmax)
        return network





from scipy.stats import pearsonr, spearmanr
import numpy as np
import os 


# #load data
# from io_modules import load_data
# #load in the data
# data_init = load_data('/projects/sjones_prj/tumourchar/data/v8/normed_sets_v8/rastminmax/v8_pogfiltered_12082016_normedrastminmax1_weightednone_keptsmallTrue.pkl.gz')
# #format data
# #test set
# x_test = data_init[2][0]

# #load reconstructed data
# #reconstructed_input = np.load('/home/jeyang/Desktop/DeepLearning/ResultsWithCV/FirstTest/cv5/reconstructions.npy')
# #reconstructed_input = np.load('/home/jeyang/Desktop/DeepLearning/Reconstructions/ChangingHiddenLayerSize/RastMinMax/reconstructedTest_deep_sigmoid_CE_drop0.0_nh_1000_ep_20.npy')
# #reconstructed_input = np.load('/projects/sjones_prj/autoencoder/reconstructedTest_sigmoid_CE_drop0.7_nh_140.npy')

# #load reconstructed data
# #switch out for file of interest
# reconstructed_input_file = '/home/jeyang/Desktop/DeepLearning/Reconstructions/ChangingHiddenLayerSize/RastMinMax/reconstructedTest_deep_sigmoid_CE_drop0.0_nh_1000_ep_20.npy'
# reconstructed_input = np.load(reconstructed_input_file)

# #create directory to output the correlations to
# output_directory_name = '/home/jeyang/Desktop/DeepLearning/Correlations/reconstructedTest_deep_sigmoid_CE_drop0.0_nh_1000_ep_20'
# if not os.path.exists(output_directory_name):
# 	os.makedirs(output_directory_name)

#get_summary(x_test, reconstructed_input)

#------------------------------- Correlation of Means (Genes) --------------------------------
#returns 2 values: pearson_coeff_str, spearman_coeff_str
def corr_of_means_genes(input_actual, input_reconstructed):
	#take means of input and reconstructed input
	#axis 0 is columns
	mean_true = np.mean(input_actual, axis=0)
	#save list of mean_true
	#np.savetxt(output_directory_name + '/mean_true_correlationOfMeans_genes', mean_true)
	mean_predicted = np.mean(input_reconstructed, axis=0)
	#save list of mean_predicted
	#np.savetxt(output_directory_name + '/mean_predicted_correlationofMeans_genes', mean_predicted)

	#Pearson Population Correlation Coefficient
	#assumes normal distribution

	#x and y inputs are each a 1-d array

	pearson_coeff = pearsonr(mean_true, mean_predicted)
	pearson_coeff_str = "pearson: " + str(pearson_coeff[0])
	print(pearson_coeff_str)
	#print("pearson: ", pearson_coeff)


	#Spearman Rank Order Correlation
	#does not assume normal distribution 

	#x and y inputs are each a 1-d array or 2-d array

	spearman_coeff = spearmanr(mean_true, mean_predicted)
	spearman_coeff_str = "spearman: " + str(spearman_coeff[0])
	print(spearman_coeff_str)
	#print("spearman: ", spearman_coeff)

	return pearson_coeff_str, spearman_coeff_str


#------------------------------- Correlation of Genes --------------------------------
#output is pearson_mean_str, pearson_variance_str, spearman_mean_str, spearman_variance_str
def corr_of_genes(input_actual, input_reconstructed):
	size = input_actual.shape[1]
	index = range(0, size)

	#Pearson Population Correlation Coefficient
	#list of correlations of each sample
	sample_correlations_pearson = []

	for i in index:
		pearson_coeff = pearsonr(input_actual[:,i],input_reconstructed[:,i])
		sample_correlations_pearson.append(pearson_coeff[0])

	#save list of sample_correlations_pearson
	#np.savetxt(output_directory_name + '/gene_correlations_pearson', sample_correlations_pearson)

	pearson_mean_genes_str = "pearson_mean: " + str(np.mean(sample_correlations_pearson))
	print(pearson_mean_genes_str)
	#print("pearson_mean: ", np.mean(sample_correlations_pearson))
	pearson_variance_genes_str = "pearson_variance: " + str(np.var(sample_correlations_pearson))
	print(pearson_variance_genes_str)
	#print("pearson_variance: ", np.var(sample_correlations_pearson))

	#Spearman Rank Order Correlation
	#list of correlations of each sample
	sample_correlations_spearman = []

	for i in index:
		spearman_coeff = spearmanr(input_actual[:,i],input_reconstructed[:,i])
		sample_correlations_spearman.append(spearman_coeff[0])

	#save list of sample_correlations_spearman
	#np.savetxt(output_directory_name + '/gene_correlations_spearman', sample_correlations_spearman)

	spearman_mean_genes_str = "spearman_mean: " + str(np.mean(sample_correlations_spearman))
	print(spearman_mean_genes_str)
	#print("spearman_mean: ", np.mean(sample_correlations_spearman))
	spearman_variance_genes_str = "spearman_variance: " + str(np.var(sample_correlations_spearman))
	print(spearman_variance_genes_str)
	#print("spearson_variance: ", np.var(sample_correlations_spearman))

	return pearson_mean_genes_str, pearson_variance_genes_str, spearman_mean_genes_str, spearman_variance_genes_str

#------------------------------- Variance of Input and Output --------------------------------
#output is average_variance_str
def variance_of_genes(input_actual, input_reconstructed):
	#take means of input and reconstructed input
	mean_true = np.mean(input_actual, axis=0)
	mean_predicted = np.mean(input_reconstructed, axis=0)

	size = input_actual.shape[1]
	index = range(0, size)

	#list of variances of each gene
	variance_per_gene = []

	#get variance of each gene
	for j in index:
		difference = abs(mean_predicted[j]-mean_true[j])
		variance_per_gene.append(difference)

	#print variance_per_gene

	average_variance = np.mean(variance_per_gene)
	average_variance_str = "average_var: " + str(average_variance)
	print(average_variance_str)
	#print("average_var: ", average_variance)

	return average_variance_str
#---------------------- Correlation of Means (Samples) ------------------------
#output is pearson_coeff_str_samples, spearman_coeff_str_samples
def corr_of_means_samples(input_actual, input_reconstructed):
	#take means of each sample (average value genes take on)
	#axis 1 is rows
	mean_true = np.mean(input_actual, axis=1)
	#save list of mean_true
	#np.savetxt(output_directory_name + '/mean_true_correlationOfMeans_samples', mean_true)
	mean_predicted = np.mean(input_reconstructed, axis=1)
	#save list of mean_predicted
	#np.savetxt(output_directory_name + '/mean_predicted_correlationOfMeans_samples', mean_predicted)

	#Pearson Population Correlation Coefficient
	#assumes normal distribution

	#x and y inputs are each a 1-d array

	pearson_coeff = pearsonr(mean_true, mean_predicted)
	pearson_coeff_str_samples = "pearson (across samples): " + str(pearson_coeff[0])
	print(pearson_coeff_str_samples)
	#print("pearson: ", pearson_coeff)


	#Spearman Rank Order Correlation
	#does not assume normal distribution 

	#x and y inputs are each a 1-d array or 2-d array

	spearman_coeff = spearmanr(mean_true, mean_predicted)
	spearman_coeff_str_samples = "spearman (across samples): " + str(spearman_coeff[0])
	print(spearman_coeff_str_samples)
	#print("spearman: ", spearman_coeff)

	return pearson_coeff_str_samples, spearman_coeff_str_samples


#------------------------- Correlation of Samples ---------------------------
#output is pearson_mean_str, pearson_variance_str, spearman_mean_str, spearman_variance_str
def corr_of_samples(input_actual, input_reconstructed):
	size = input_actual.shape[0]
	index = range(0, size)

	#Pearson Population Correlation Coefficient
	#list of correlations of each sample
	sample_correlations_pearson = []

	for i in index:
		pearson_coeff = pearsonr(input_actual[i,:],input_reconstructed[i,:])
		sample_correlations_pearson.append(pearson_coeff[0])

	#save list of sample_correlations_pearson
	#np.savetxt(output_directory_name + '/sample_correlations_pearson', sample_correlations_pearson)

	pearson_mean_samples_str = "pearson_mean (across samples): " + str(np.mean(sample_correlations_pearson))
	print(pearson_mean_samples_str)
	#print("pearson_mean: ", np.mean(sample_correlations_pearson))
	pearson_variance_samples_str = "pearson_variance (across samples): " + str(np.var(sample_correlations_pearson))
	print(pearson_variance_samples_str)
	#print("pearson_variance: ", np.var(sample_correlations_pearson))

	#Spearman Rank Order Correlation
	#list of correlations of each sample
	sample_correlations_spearman = []

	for i in index:
		spearman_coeff = spearmanr(input_actual[i,:],input_reconstructed[i,:])
		sample_correlations_spearman.append(spearman_coeff[0])

	#save list of sample_correlations_spearman
	#np.savetxt(output_directory_name + '/sample_correlations_spearman', sample_correlations_spearman)

	spearman_mean_samples_str = "spearman_mean (across samples): " + str(np.mean(sample_correlations_spearman))
	print(spearman_mean_samples_str)
	#print("spearman_mean: ", np.mean(sample_correlations_spearman))
	spearman_variance_samples_str = "spearman_variance (across samples): " + str(np.var(sample_correlations_spearman))
	print(spearman_variance_samples_str)
	#print("spearson_variance: ", np.var(sample_correlations_spearman))

	return pearson_mean_samples_str, pearson_variance_samples_str, spearman_mean_samples_str, spearman_variance_samples_str


#------------------------------ List of Values -------------------------------
#get summary of all correlation values
def get_summary(input_actual, input_reconstructed, output_directory_name):
#save correlations to text file
	pearson_coeff_str, spearman_coeff_str = corr_of_means_genes(input_actual, input_reconstructed)
	pearson_mean_genes_str, pearson_variance_genes_str, spearman_mean_genes_str, spearman_variance_genes_str = corr_of_genes(input_actual, input_reconstructed)
	average_variance_str = variance_of_genes(input_actual, input_reconstructed)
	pearson_coeff_str_samples, spearman_coeff_str_samples = corr_of_means_samples(input_actual, input_reconstructed)
	pearson_mean_samples_str, pearson_variance_samples_str, spearman_mean_samples_str, spearman_variance_samples_str = corr_of_samples(input_actual, input_reconstructed)
	list_of_values = [pearson_coeff_str, spearman_coeff_str, pearson_mean_genes_str, pearson_variance_genes_str, spearman_mean_genes_str, spearman_variance_genes_str, 
		average_variance_str, pearson_coeff_str_samples, spearman_coeff_str_samples, pearson_mean_samples_str, pearson_variance_samples_str, 
		spearman_mean_samples_str, spearman_variance_samples_str]
	#list_of_values = [pearson_coeff_str, spearman_coeff_str]
	print(list_of_values)
	#save list of all values
	output_file_name = output_directory_name + '/summaryOfAllCorrelations.txt'

	file = open(output_file_name, 'w')

	for value in list_of_values:
		file.write("%s\n" % value)


#get_summary(x_test, reconstructed_input, output_directory_name)


#--------------------------- Correlations 2d --------------------------------
#A is the actual input, B is the reconstructed input
def corr_2d(A,B):	#A is a datafrmae, B is a dataframe
	A_mA = A - A.mean(1)[:,None]
	B_mB = B - B.mean(1)[:,None]
	ssA = (A_mA**2).sum(1);
	ssB = (B_mB**2).sum(1);
	temp_corr = np.dot(A_mA,B_mB.T)/np.sqrt(np.dot(ssA[:,None],ssB[None]))
	mean_corr = np.diagonal(temp_corr).mean()
	
	return np.dot(A_mA,B_mB.T)/np.sqrt(np.dot(ssA[:,None],ssB[None]))



#Contains commonly used classes and functions
#For I/O purposes in ML model runs

#Shared libs
import six.moves.cPickle as pickle
import gzip
import os
import sys
import timeit
import glob 
import numpy as np
import pandas as pd
import theano
import theano.tensor as T

from argtypes import Dictionary
from sampling_algorithms import *

import datetime
from collections import Counter
import random
import itertools
import copy
from sklearn.model_selection import StratifiedKFold
from sklearn.utils import shuffle
from split_input_into_testtrainvalid import *

#Definitions
import logging, sys

def subset_features_x(x, fullfeatures_named, subsetfeatures_named, fillzero=False):
        #Used to select a set of input features in a specific order
        #Format of input is samples x features
        #Order of features is fullfeatures_named; subsetting is subsetfeatures_named
        subset_indices = []; missing_values=[]
        for select_this in subsetfeatures_named:
                subset_indices.append(fullfeatures_named.index(select_this))
        #returned object will have the features ordered in the manner of subsetfeatures_named
        return (np.apply_along_axis(lambda x: x[subset_indices], 1, x))

class Unbuffered3:
	def __init__(self, stream):
		root = logging.getLogger()
		root.setLevel(logging.DEBUG)
		ch = logging.StreamHandler(sys.stdout)
		ch.setLevel(logging.DEBUG)
		formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
		ch.setFormatter(formatter)
		root.addHandler(ch)


class Unbuffered:
        def __init__(self, stream, logfile=open("logfile.txt",'w')):
                self.stream = stream
		self.logfile = logfile
		print("writing to {0}".format(logfile))	
	def flush(self):
		pass        
        def write(self, data):
                self.stream.write(data)
		self.stream.flush()
                self.logfile.write(data)

def load_data(dataset, make_shared=False):
    ''' Loads the dataset

    :type dataset: string
    :param dataset: the path to the dataset (here MNIST)
    '''

    data_dir, data_file = os.path.split(dataset)
    if data_dir == "" and not os.path.isfile(dataset):
        # Check if dataset is in the data directory.
        new_path = os.path.join(
            os.path.split(__file__)[0],
            "..",
            "data",
            dataset
        )
        if os.path.isfile(new_path):
            dataset = new_path

    print('... loading data from {0}'.format(dataset))
    # Load the dataset
    with gzip.open(dataset, 'rb') as f:
        try:
            train_set, valid_set, test_set = pickle.load(f, encoding='latin1')
        except:
            train_set, valid_set, test_set = pickle.load(f)
    # train_set, valid_set, test_set format: tuple(input, target)
    # input is a numpy.ndarray of 2 dimensions (a matrix)
    # where each row corresponds to an example. target is a
    # numpy.ndarray of 1 dimension (vector) that has the same length as
    # the number of rows in the input. It should give the target
    # to the example with the same index in the input.
    def shared_dataset(data_xy, borrow=True):
        """ Function that loads the dataset into shared variables

        The reason we store our dataset in shared variables is to allow
        Theano to copy it into the GPU memory (when code is run on GPU).
        Since copying data into the GPU is slow, copying a minibatch everytime
        is needed (the default behaviour if the data is not in a shared
        variable) would lead to a large decrease in performance.
        """
        data_x, data_y = data_xy
        shared_x = theano.shared(numpy.asarray(data_x,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        shared_y = theano.shared(numpy.asarray(data_y,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        # When storing data on the GPU it has to be stored as floats
        # therefore we will store the labels as ``floatX`` as well
        # (``shared_y`` does exactly that). But during our computations
        # we need them as ints (we use labels as index, and if they are
        # floats it doesn't make sense) therefore instead of returning
        # ``shared_y`` we will have to cast it to int. This little hack
        # lets ous get around this issue
        return shared_x, T.cast(shared_y, 'int32')

    if make_shared:
        test_set_x, test_set_y = shared_dataset(test_set)
        valid_set_x, valid_set_y = shared_dataset(valid_set)
        train_set_x, train_set_y = shared_dataset(train_set)

        rval = [(train_set_x, train_set_y), (valid_set_x, valid_set_y),
            (test_set_x, test_set_y)]
    else:
        rval = [train_set, valid_set, test_set]
    return rval

def load_cpickled_data(dataset):
        '''Load the data. Presume it is cPickled [(trainx,trainy),(validx,validy),(testx,testy)]
        	- Currently imported by SVM models
	'''
        print("...loading data")
        with gzip.open(dataset, 'rb') as f:
                try:
                        train_set, valid_set, test_set = pickle.load(f, encoding='latin1')
                except:
                        train_set, valid_set, test_set = pickle.load(f)

        train_set_x, train_set_y = train_set
        valid_set_x, valid_set_y = valid_set
        train_set_x = np.concatenate((train_set_x, valid_set_x), axis=0)
        train_set_y = np.asarray(list(train_set_y) + list(valid_set_y))
        test_set_x, test_set_y = test_set
        rval = [(train_set_x, train_set_y), (test_set_x, test_set_y)]
        return rval

def load_sig_dat(infile, prefix, outdir, testpog=False):
        t1 = pd.read_table(infile)
        if prefix == "exposures":
                cols = ["id","tumour_type","signature","exposure_per_mb"]
                featname = "signature"; featvalue = "exposure_per_mb"
        elif prefix == "catalogs":
                cols = ["id", "tumour_type", "mutation_type", "mutations_per_mb"]
                featname = "mutation_type"; featvalue="mutations_per_mb"
	else:
		return(0)
        if testpog:
                t1=t1[t1.sample_prefix.str.contains("biop")==True]
                t1['id'] = t1[['id', 't','tumour_library','n','normal_library']].apply(lambda x: '_'.join(x), axis=1)
                t1['tumour_type'] = "unknown"
                prefix = prefix + "_pog"
        tdat=t1[cols]
        tdat_feats = tdat.pivot_table(index=["id","tumour_type"], columns=[featname], values=featvalue)
        tdat_feats = tdat_feats.reset_index().rename_axis(None,axis=1)
        intxt = outdir + "/data_" + prefix + ".txt"
        insamples = outdir + "/samples_" + prefix + ".txt"
        tdat_feats.ix[:,0:2].to_csv(insamples, index=None, sep="\t", mode="w")
        tdat_feats.ix[:,1:].to_csv(intxt, index=None, sep='\t', mode="w")
        return([tdat_feats, intxt, insamples])





def readtestinput(datfile, trimlabs, header_status, sep='\t'):
        "This function takes in a matrix of format genes(columns) x samples(rows) and generates an array of labels (Y) and an array of corresponding input vectors (X). The first column is the sample names/labels - Currently imported by SVM models"
        with open(datfile, 'r') as infile:
                reader = csv.reader(infile, delimiter=sep)
                X = []
                Y = []
                labels = []
                samples = []
                header = header_status
                for row in reader:
                        if header:
                                header = False
                                continue
                        row[2:] = [float('nan') if elem=="NA" else float(elem) for elem in row[2:]]
                        X.append(row[2:])
                        Y.append("0")
                        labels.append(row[0])
                        samples.append(row[1])
                X = np.asarray(X)
                #Y = np.asarray(Y)
                #Y = Y.reshape(Y.shape[0],1)
                #Update labels (from type_snum to type, and then a list of numeric labels for output types)  
                if trimlabs:
                        labels = [re.sub('[0-9]*','',x) for x in labels]
	return [X, Y, labels, samples]

def load_txt_input(datfile, trimlabs, header_status, drugbinderset, sep = '\t'):
    """This function takes in a matrix of format genes(columns) x samples(rows) and generates an array of labels (Y) and an array of corresponding input vectors (X). The first column is the sample names/labels """
    with open(datfile, 'r') as infile:
        reader = csv.reader(infile, delimiter=sep)
        X = []
        Y = []
        labels = []
        header = header_status
        if drugbinderset:
            ligands = []
            for row in reader:
                if header:
                    header = False
                    continue
                row[2:] = [ (float('nan') if elem == 'NA' else float(elem)) for elem in row[2:] ]
                X.append(row[2:])
                ligands.append(row[0])
                Y.append(row[1])
                labels.append(row[1])

            ligands = np.asarray(ligands)
            X = np.asarray(X)
            Y = np.asarray(Y)
            Y = Y.reshape(Y.shape[0], 1)
            if trimlabs:
                labels = [ re.sub('[0-9]*', '', x) for x in labels ]
            returnset = [X,
             Y,
             labels,
             ligands]
        else:
            for row in reader:
                if header:
                    header = False
                    continue
                row[1:] = [ (float('nan') if elem == 'NA' else float(elem)) for elem in row[1:] ]
                X.append(row[1:])
                Y.append(row[0])
                labels.append(row[0])

            X = np.asarray(X)
            Y = np.asarray(Y)
            Y = Y.reshape(Y.shape[0], 1)
            if trimlabs:
                labels = [ re.sub('[0-9]*', '', x) for x in labels ]
            returnset = [X,Y, labels,0]
        return returnset

class mldataset(object):
        def __init__(self, dataset_file, outdir, normtype="none",misclassflag="none", normby=str(1), keepsmall=True, deeplearn=False):
                #Provide input address
                self.inputfile = dataset_file
                self.outdir = outdir
                self.keepsmall = keepsmall
                self.normtype = normtype
                self.normby = normby
		self.shuffle_train = True
		self.misclassflag = misclassflag	
		self.deeplearn = deeplearn
		
        def load_mldata(self, fullset=False, verbose=True, save_full_stats=False, subsetfeatures=None):
		#subsetfeatures must be a txt file with list fo features to subset input with
                if(self.inputfile.endswith(".tsv")|self.inputfile.endswith(".txt")):
                        self.load_txtdata(fullset=fullset, verbose=verbose, save_full_stats=save_full_stats)
                else:
                        self.load_pkldata()
		self.X = np.concatenate((self.datasets[0][0], self.datasets[1][0], self.datasets[2][0]))
		self.y = np.concatenate((self.datasets[0][1], self.datasets[1][1], self.datasets[2][1]))
		
		if subsetfeatures is not None:
			if verbose:
				print("SUBSETTING TRAINING DATA FEATURES from {0}".format(subsetfeatures))
			features_to_subset = [line.rstrip().strip('"') for c, line in enumerate(open(subsetfeatures[0], 'r'))]
			self.X = subset_features_x(self.X, self.features, features_to_subset, fillzero=False)
			self.features = features_to_subset
		
		if save_full_stats:
			#Split data into train, validation, and test
			self.getCVfolds(self, k=5, validation_size = 0.1, makeupdate=True)
						
        def load_txtdata(self,fullset=False,verbose=True, save_full_stats=False):
                #This part splits and normalizes the data
                dataseparator = mydatstuff(infile=self.inputfile, outdir=self.outdir, trimflag=True, keepheader=True, drugbinderflag=False,keepsmall=self.keepsmall, fulltrainflag=fullset, classligands=None, testsize=0, outprefix="prepared_dat", normtype=self.normtype, normby=str(self.normby), makevalid=True,verbose=verbose, deeplearn=self.deeplearn)
                dataseparator.splitvalidtest()
                dataseparator.normalize()
                self.datasets, self.labelnames, self.labweighted, self.features = dataseparator.return_dataset()
                if save_full_stats:
                        dataseparator.writeout() #Save idx's, not the actual dataset
			self.save_preprocessed_files(verbose=verbose)
		
        def load_pkldata(self):
                #Assume pickled datasets are already normalized
                #Pickled datasets are also in 3 sets, train-validation-test (3 idxs)

                self.datasets = load_data(self.inputfile, make_shared=False)
                indir = os.path.dirname(os.path.abspath(self.inputfile))
		self.labelweights = glob.glob(indir+"/dict_labweights_*")[0]
                self.labelset = glob.glob(indir+"/dict_labels_*")[0]
                self.featurelist = glob.glob(indir+"/featurelist.txt")
                self.labweighted = Dictionary(labelset=self.labelweights)
                self.labelnames = Dictionary(labelset=self.labelset)
                
		if self.featurelist:
                        f = open(self.featurelist[0])
                        with open(self.featurelist[0]) as f:
                                self.features = [line.rstrip() for line in f]

        def getCVfolds(self, k=2, validation_size = 0, X = None, y=None, makeupdate=False):
		k = int(k)
		validation_size= float(validation_size)
		if X is None:
			X = self.X
			y = self.y
                if k==1:
			k = 2
			#Default CV folds are 2 (for a single run, select first return)
                
		#First remove the samples and classes that are represented at freq < k (disrupts folds)
                class_counts = Counter(y)
                dropclass = [lk for lk in dict(class_counts).keys() if class_counts[lk] < k]
                smallidx = [np.where(y==f)[0] for f in dropclass]
                ##print("-------\nDROPCLASSES: {0} ".format(dropclass))
		#Stratified K fold splitter. Returns k sets of [train, test] idx's
                skf = StratifiedKFold(n_splits=int(k))
                train_klist = []; test_klist = []
                
		#Get the k-fold set of train and test indices, and also append a random subset of samples from the small samples set
                if len(smallidx) > 0:
			print("--Appending small classes separately {0}".format(dropclass))
			largeclass = [lk for lk in dict(class_counts).keys() if class_counts[lk] >= k]
			largeidx = [np.where(y==f)[0] for f in largeclass]
			largeidx_list = list(itertools.chain.from_iterable(largeidx)) 
			Xl = np.take(X, largeidx_list, axis=0) 
			yl = np.take(y, largeidx_list, axis=0)
			
			skf.get_n_splits(X,y)
			
			for train_idx, test_idx in skf.split(X,y):
				#Remove any of the small indices
				train_idx = [mk for mk in train_idx if not mk in list(itertools.chain.from_iterable(smallidx))]
				test_idx = [mk for mk in test_idx if not mk in list(itertools.chain.from_iterable(smallidx))]
				
				#Randomly select the small indices from the separate list to backfill for the train and test segs
				smalltest_k = [random.choice(m) for m in smallidx]
                        	#print("{0} and contents {1} \n -----> ".format(len(smallidx),smalltest_k, np.take(y, smalltest_k,axis=0)))
				smalltrain_k = [mx for m in smallidx for mx in m if mx not in smalltest_k]
				
				#print("SUBSET HERE: \n{0} \n AND \n{1}\nAND TEST SUBSET{2}\nAND TEST {3}\n".format(train_idx,smalltrain_k, test_idx,smalltest_k))
				train_klist.extend([np.concatenate((train_idx, smalltrain_k))])
                        	test_klist.extend([np.concatenate((test_idx,smalltest_k))])
		else:
			skf.get_n_splits(X,y)
			for train_idx, test_idx in skf.split(X,y):
				train_klist.extend([train_idx])
				test_klist.extend([test_idx])
		
		if makeupdate:
			print("...Updated K fold sets for training and test.")
			self.trainidx_K = train_klist	
			self.testidx_K = test_klist
		
		if validation_size > 0.0 and validation_size <= 0.5 :
			self.trainvalididx_K = [] ##self.trainidx_K #train_klist
			for i in range(0,len(train_klist)):
				bigtrain_idx = train_klist[i].astype(int)
				X = np.take(self.X , bigtrain_idx, axis=0)
				train_kth, valid_kth = self.getCVfolds(k=int(1/validation_size), X = np.take(self.X , bigtrain_idx, axis=0), y=np.take(self.y, bigtrain_idx, axis=0))
				self.trainvalididx_K.extend([[train_kth[0], valid_kth[0]]])
		
		return(train_klist, test_klist)
		
	def save_preprocessed_files(self, verbose=True):
                if(self.inputfile.endswith(".txt")):
                        #If the input file was a txt, only then do you save preprocessed files (dont repickled pickled input, pointless)
                        #Create an output data directory
                        try:
                                self.datadir
                        except:
                                self.datadir=self.outdir + "/preprocessed_data"
                        try:
                                os.stat(self.datadir)
                        except:
                                os.mkdir(self.datadir)
			if verbose:
                        	print("...Pickling preprocessed, split data to output directory {0}".format(self.datadir))
                        filename = os.path.splitext(os.path.basename(self.inputfile.rsplit('.', 1)[0]))[0]
                        self.dataoutpref = filename + "_normed" + self.normtype + str(self.normby) + "_keptsmall" + str(self.keepsmall)
                        outall = gzip.open(os.path.join(self.datadir, ''.join([self.dataoutpref, '.pkl.gz'])), 'wb')
			
			
                        cPickle.dump([[self.X_train, self.y_train], [self.X_val, self.y_val], [self.X_test, self.y_test]], outall, protocol=cPickle.HIGHEST_PROTOCOL)
                        if verbose:
				print("   - pkl dump of split data at {0}".format(os.path.join(self.datadir, ''.join([self.dataoutpref, '.pkl.gz']))))
                        with open(os.path.join(self.datadir, ''.join(['dict_labels_', self.dataoutpref, '.txt'])), 'w') as f:
                                [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labelnames.items() ]
                        with open(os.path.join(self.datadir, ''.join(['dict_labweights_', self.dataoutpref, '.txt'])), 'w') as f:
                                [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labweighted.items() ]
                        with open(os.path.join(self.datadir, 'featurelist.txt'), 'w') as f:
                                [ f.write('{0}\n'.format(x)) for x in self.features]

                        if verbose:
				print("...Updated input file")

                        self.inputfile = os.path.join(self.datadir, ''.join([self.dataoutpref, '.pkl.gz']))
                else:
			if verbose:
                        	print("...Not pickling data because input was a pkl'd split dataset already")
	
	def get_twosets(self, set1idx, set2idx):
                X1 = np.take(self.X, set1idx, axis=0)
                y1 = np.take(self.y, set1idx, axis=0)
                X2 = np.take(self.X, set2idx, axis=0)
                y2 = np.take(self.y, set2idx, axis=0)
                return([[X1,y1],[X2,y2]])

	def save_dataset_details(self,outdir, write_idx=True):
		print("====OUTDIR = {0}".format(self.outdir))
		with open(''.join([self.outdir,"/featurelist.txt"]), 'w') as f:
			[ f.write('{0}\n'.format(m)) for m in self.features ]
		
		if(self.inputfile is None):
			filename="predefined_input"
		else:
			filename=os.path.splitext(os.path.basename(self.inputfile.rsplit('.', 1)[0]))[0]
		self.outpref = filename	
		self.datoutdir = os.path.join(os.path.abspath(self.outdir) + '/dataset_train/' + filename + '_' + datetime.datetime.now().strftime('%Y-%m-%d'))
		print("Saving Dataset details to {0}".format(self.datoutdir))
		if not os.path.exists(self.datoutdir):
			os.makedirs(self.datoutdir)
		 
		if write_idx:
			print("Writing idx's to output, NOTE if this is full training then test and train idxs are combined for train. Validation becomes test set too")
			with open(os.path.join(self.datoutdir, 'trainingidxs.txt'), 'w') as f:
				[ f.write('{0}\n'.format(x)) for x in self.sub_kfold_trainidx ]
			with open(os.path.join(self.datoutdir, 'testidxs.txt'), 'w') as f:
				[ f.write('{0}\n'.format(x)) for x in self.sub_kfold_testidx ]
			with open(os.path.join(self.datoutdir, 'valididxs.txt'), 'w') as f:
				[ f.write('{0}\n'.format(x)) for x in self.sub_kfold_valididx ]
		
		try:
			int(self.labelnames.keys()[0])
		except ValueError:
			self.labelnames = {k:v for v,k in self.labelnames.iteritems()}
				
		with open(os.path.join(self.datoutdir, ''.join(['dict_labels_', self.outpref, '.txt'])), 'w') as f:
            		[ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labelnames.items() ]
        	with open(os.path.join(self.datoutdir, ''.join(['dict_labcount_', self.outpref, '.txt'])), 'w') as f:
         		[ f.write('{0}\t{1}\n'.format(self.labelnames[key], value)) for key, value in dict(Counter(self.y)).items() ]   
	#		[ f.write('{0}\t{1}\n'.format(key, value)) for key, value in dict(Counter(np.concatenate((self.y_train, self.y_val, self.y_test),axis=0))).items() ]
		
	        with open(os.path.join(self.datoutdir, ''.join(['dict_labweights_', self.outpref, '.txt'])), 'w') as f:
        	    [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labweighted.items() ]
		
	def preprocess_data(self, kfold, features_subset=[], verbose=True, fullset=False):
		if '0' in self.labelnames.keys() or 0 in self.labelnames.keys():
                        self.labelnames = dict([int(a),b] for a,b in self.labelnames.iteritems())
                else:
			if self.deeplearn is True:
				self.labelnames = dict([b,a] for a,b in self.labelnames.iteritems())
			else:
				self.labelnames = dict([int(b),a] for a,b in self.labelnames.iteritems())
		
                self.sorted_keylabs=[]
                for i in sorted(self.labelnames.keys()):
                                self.sorted_keylabs.append(self.labelnames[i])
		
		main_trainidx = self.trainidx_K[kfold-1]
		
		self.sub_kfold_trainidx, self.sub_kfold_valididx = self.trainvalididx_K[kfold-1]
		self.sub_kfold_testidx = self.testidx_K[kfold-1]
		
		kfold_trainidx = np.take(main_trainidx, self.sub_kfold_trainidx); 
		kfold_valididx = np.take(main_trainidx, self.sub_kfold_valididx)
		kfold_testidx = self.sub_kfold_testidx		
		
		kfold_trainidx = kfold_trainidx.astype(int)
		kfold_valididx = kfold_valididx.astype(int)
		kfold_testidx = kfold_testidx.astype(int)
		
		if fullset is True:
			if verbose:
				print("...Appending Test idxs to training prior to sample inflation \n\t Validation set is test for full model training")
			kfold_trainidx = np.concatenate((kfold_trainidx, kfold_testidx),axis=0)
		#	kfold_trainidx = list(set(kfold_trainidx + kfold_testidx))
			kfold_testidx = kfold_valididx
			
                [[self.X_train, self.y_train], [self.X_val, self.y_val]] = self.get_twosets(set1idx = kfold_trainidx, set2idx = kfold_valididx)
		[[self.X_traintemp, self.y_traintemp], [self.X_test, self.y_test]] = self.get_twosets(set1idx = kfold_trainidx, set2idx = kfold_testidx)
		
		if verbose:
			print("Train shape {0} \nValid shape {1} \nTest shape {2}".format(self.y_train.shape,self.y_val.shape, self.y_test.shape))
			print("X shape {0} \nY shape {1}".format(self.X.shape,self.y.shape))
			print("Train + valid uniq {0} \n{1}, overall uniq {2}".format(len(np.unique(self.trainidx_K[kfold-1])),np.unique(np.concatenate((kfold_valididx,kfold_trainidx),axis=0)).shape,np.unique(np.concatenate((kfold_testidx,kfold_valididx,kfold_trainidx),axis=0)).shape))	
			print("{0} -- {1} -- {2}".format(len(set(self.y_train)), len(set(self.y_val)), len(set(self.y_test))))	
			#if np.array_equal(set(kfold_trainidx + kfold_valididx + kfold_testidx), self.y.shape[0]):
			#	del self.X_traintemp; del self.y_traintemp
			#else:
			#	print("ERROR: There is something wrong with your K fold split, the Train fold is not consistent between validation and test pairs")
	
			print(set(self.y_val))	
                	print(self.X_train.shape, self.X_val.shape, self.X_test.shape)
                
		if len(features_subset) > 0:
                        self.X_train = subset_features_x(self.X_train, self.features, features_subset)
                        self.X_val = subset_features_x(self.X_val, self.features, features_subset)
                        self.X_test = subset_features_x(self.X_test, self.features, features_subset)
                        if verbose:
				print("___SUBSET INPUT, new training data shape {0}".format(self.X_train.shape))
                        with open(os.path.join(self.outdir, 'featurelist.txt'), 'w') as f:
                                [ f.write('{0}\n'.format(x)) for x in features_subset]
		
                self.n_class = len(set(self.y_train)) #Training set represents all the classes
                self.n_in = len(self.X_train[0]) #First input sample's length is the number of features (assumes it remains the same for all samples!)

                if isinstance(self.y_test[0], np.int):
                        if verbose:
				print("...--Verified test is correctly labelled--...")
                else:
			if not self.deeplearn:	
                        	if verbose:
					print("...--Test labels are alphabetical, mapping to numeric labels--...")
                        	#Temporarily invert labelnames dict so key:value is labelname:labelnum
                        	invert_dict = dict([b,a] for a,b in self.labelnames.iteritems())
                        	numeric_test_real = [0] * len(self.y_test)
                        	for i in range(self.y_test.shape[0]):
                                	if self.y_test[i] == "COAD_TS":
                                        	self.y_test[i] = "COADREAD_TS"
                                	numeric_test_real[i] = invert_dict[self.y_test[i]]
                        	self.y_test = np.asarray(numeric_test_real)

                if len(set(self.y_train)) != len(set(self.y_test)):
                        if verbose:
				print("WARNING: Test set does not have all the classes from the training set")
                        	print(set(self.y_train), set(self.y_test))
                self.save_preprocessed_files(verbose=verbose)

                if self.misclassflag == "none":
                        if verbose:
				print("No synthetic additions/modifications to the data")
                elif self.misclassflag != "weight":
                        self.X_train, self.y_train = adjust_trainset(self.X_train, self.y_train, self.misclassflag, verbose=verbose)
                        self.X_train, self.y_train  = shuffle(self.X_train, self.y_train, random_state=0) #Must reshuffle after synthetic additions

                if(self.shuffle_train):
                        if verbose:
				print("...Shuffling training set")
                        self.X_train, self.y_train  = shuffle(self.X_train, self.y_train, random_state=0)
                self.X_train = np.asarray(self.X_train)
                self.y_train = np.asarray(self.y_train)	
	



import matplotlib
matplotlib.use('Agg')
#from plotnine import *
#from pandas.api.types import CategoricalDtype
from .config import REF_LABELCOUNTS, REF_DISEASECODES
import glob
import re
import pandas as pd
import numpy as np
import os, sys
#### Get class sizes  
labelsize_df = pd.read_csv(REF_LABELCOUNTS, delimiter="\t", header=None)
labelsize_df.columns = ["classlabel", "class_size"]
labelsize_df["class_size_rescaled"] = labelsize_df["class_size"].astype(float)/labelsize_df["class_size"].astype(float).max()
labelsize_df['label_sorted'] = [re.split('_[T,N]S$', m)[0] for m in  labelsize_df.classlabel.tolist()]
labelsize_df['tissuestatus'] = ["Tumor" if m is True else "Normal" for m in ['_TS' in m for m in  labelsize_df.classlabel.tolist()]]
labelsize_df['pred'] = labelsize_df['class_size_rescaled']
labelsize_df['Source'] = None

#### Get predictions across all cases, categories, models 
codes_disease = pd.read_csv(REF_DISEASECODES, delimiter="\t")

#### Function to make plot-ready dataframe from ensemble prediction dictionary (each model = each key, respective value is a 3d tensor)
def get_plotting_df(dict_with_preds, x_sample_names=None):
        modelnames = list(dict_with_preds.keys())
        df_temp_matrix = np.array([dict_with_preds[i] for i in modelnames])
        num_models, num_samples, num_preds, num_tuples_preds = df_temp_matrix.shape
        flat_matrix = df_temp_matrix.flatten()
        for i in range(0, num_models):
                for j in range(0, num_samples):
                        if i == 0 and j == 0:
                                ret_df = pd.DataFrame(df_temp_matrix[i, j]); ret_df["model"] = modelnames[i]; ret_df["sample_ix"] = j
                        else:
                                new_data_line = pd.DataFrame(df_temp_matrix[i, j]); new_data_line["model"] = modelnames[i]; new_data_line["sample_ix"] = j
                                ret_df = ret_df.append(new_data_line)
        ret_df.columns = ["classlabel", "pred", "model", "sample_ix"]
        if x_sample_names is not None:
                ret_df['sample_name'] = [x_sample_names[m] for m in ret_df['sample_ix'].tolist()]
        ret_df['tissuestatus'] = ["Tumor" if m is True else "Normal" for m in ['_TS' in m for m in  ret_df.classlabel.tolist()]]
        ret_df['label'] = [re.split('_[T,N]S$', m)[0] for m in  ret_df.classlabel.tolist()]
        return ret_df


#### Function to actually plot!  
def plot_cases(ensembledf_with_samples, plot_df_with_samples, outdir, save_txt=False):
	### ensembledf_with_samples is received as output from scope_ensemble.get_ensemble_score()
	### plot_df_with_samples is received as output from scope_plots.get_plotting_df()
	### Needs global dfs labelsize_df, codes_disease
	global labelsize_df, codes_disease
	## First check if sample names provided, otherwise use sample indices as names
	if 'sample_name' not in ensembledf_with_samples:
		sys.stdout.write("\nNo sample names provided, using sample indices\n")
		ensembledf_with_samples['sample_name'] = ensembledf_with_samples['sample_ix']
		plot_df_with_samples['sample_name'] = plot_df_with_samples['sample_ix']
	elif 'sample_name' not in plot_df_with_samples:
		sys.stdout.write("\nNo sample names provided, using sample indices\n")
		ensembledf_with_samples['sample_name'] = ensembledf_with_samples['sample_ix']
		plot_df_with_samples['sample_name'] = plot_df_with_samples['sample_ix']
	#### Get top ranked predictions for each sample
	ensembledf_with_samples_rank1 = ensembledf_with_samples[ensembledf_with_samples.rank_pred ==1]
	ensembledf_with_samples_rank1['label_sorted'] = [re.split('_[T,N]S$', m)[0] for m in ensembledf_with_samples_rank1.label.tolist()]
	ensembledf_with_samples_rank1['tissuestatus'] = ["Tumor" if m is True else "Normal" for m in ['_TS' in m for m in ensembledf_with_samples_rank1.label.tolist()]]
	#### Prepare merged plot df
	plot_df_with_samples['pred'] = plot_df_with_samples.pred.astype(float)
	df_plt = pd.merge(plot_df_with_samples, codes_disease, how='left', left_on='label', right_on='Code Name')
	df_plt = df_plt.sort_values(['Source', 'classlabel'], ascending=[1,0]).reset_index()
	label_list = df_plt['label'].unique().tolist()
	#label_cat = pd.api.types.CategoricalDtype(categories = label_list, ordered=True) ## UPDATED March 17 2020
	#label_cat = pd.CategoricalDtype(categories = label_list, ordered=True)
	#df_plt['label_sorted'] = df_plt['label'].astype(str).astype(label_cat)
	if save_txt is True:
		df_plt.to_csv(outdir + "/SCOPE_allPredictions.txt", sep="\t", index=False)
		ensembledf_with_samples.to_csv(outdir + "/SCOPE_topPredictions.txt", sep="\t", index=False)
	return




import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as pyplot
import numpy
def write_stats(outdir, pc, outdict,keyset, outname,title):
        #Sort by key alphabetical order
        valueset = []
        #keyset = sorted(outdict.keys())
        for key in keyset:
                if key in outdict:
                        valueset.append(outdict[key])

        with open(outdir + "/" + outname + ".txt", 'w') as f_out:
                for key in keyset:
			if key in outdict: 
                        	f_out.write('{0}\t{1}\n'.format(key,outdict[key]))

	#Plot sorted set
        pyplot.figure(pc)
        index = numpy.arange(len(outdict))
        pyplot.xticks(rotation=90)
	barwidth=1; pyplot.bar(range(len(valueset)), valueset, barwidth, alpha=0.8, align='center')
        pyplot.grid(); pyplot.title(title); pyplot.xticks(index, keyset, size='small')
        pyplot.tick_params(axis='both', which='major', labelsize=5.5)
        pyplot.ylabel(outname); pyplot.ylim(1e-4,1); pyplot.savefig(outdir+"/"+outname+".png",facecolor='w')

def calculate_precision(true_labs, pred_labs):
	test_classes = set(list(pred_labs) + list(true_labs))	
	#test_classes = set(pred_labs + true_labs)
        correct_counts = {c: 0 for c in test_classes}
        total_counts = {c: 0 for c in test_classes}
        precision = {c: 0 for c in test_classes}
        for sample in xrange(len(true_labs)):
                actual_class = str(true_labs[sample])
                predicted_class = str(pred_labs[sample])
                if actual_class == predicted_class:
                        correct_counts[actual_class] += 1
                        total_counts[actual_class] += 1
                else:
                        total_counts[predicted_class] += 1

        for c in correct_counts:
                precision[c] = (correct_counts[c] * 1.0) / max(1.0, total_counts[c] * 1.0)

        return(precision)

def calculate_recall(true_labs, pred_labs):
	test_classes = set(list(pred_labs) + list(true_labs))
        #test_classes = pred_labs + true_labs
        correct_counts = {c: 0 for c in test_classes}
        total_counts = {c: 0 for c in test_classes}
        recall = {c: 0 for c in test_classes}
        for sample in xrange(len(true_labs)):
                actual_class = str(true_labs[sample])
                predicted_class = str(pred_labs[sample])
                if actual_class == predicted_class:
                        correct_counts[actual_class] += 1
                        total_counts[actual_class] += 1
                else:
                        total_counts[actual_class] += 1

        for c in correct_counts:
                recall[c] = (correct_counts[c] * 1.0) / max(1.0, total_counts[c] * 1.0)

        return(recall)

def calculate_fbeta_score(true_labs, precision, recall,pred_labs):
        test_classes = set(list(pred_labs) + list(true_labs))
	#test_classes = pred_labs + true_labs
        correct_counts  = {c: 0 for c in test_classes}
        fbeta = {c: 0 for c in test_classes}
        for c in correct_counts:
                fbeta[c] = 2.0 * ((precision[c] * recall[c]) / max(1.0, (precision[c] + recall[c])))

        return(fbeta)




import lasagne
import numpy as np
#Define implicit matrix for input using a theano fuction
from theano import function
import theano.tensor as T
import theano

#Input is a 1 dimensional vector of all variabls
x = T.vector('data_vector', dtype='float32')
keep_idx = T.vector('norm_idx', dtype='int8')
norm_axis = x[keep_idx]
norm_axis_no0inv = T.switch(theano.tensor.eq(norm_axis,0), 1, 1/norm_axis)
opp_axis = x
return_mat = theano.tensor.dot(norm_axis_no0inv.reshape((-1,1)), opp_axis.reshape((1,-1)))
fn_make_mat_theano = theano.function(inputs = [x, keep_idx], outputs=return_mat)

#Set 0 values in normalizing vector to 1; flip/invert others to make them scaling factors
class vec_to_image(lasagne.layers.Layer):
	def __init__(self, incoming, num_x1, num_x2, norm_idx, **kwargs):
		super(vec_to_image, self).__init__(incoming,  **kwargs)
		self.num_x1 = num_x1
		self.num_x2 = num_x2
		self.norm_idx = norm_idx
	
	def get_output_for(self, input,**kwargs):
		#return fn_make_mat_theano(input, self.norm_idx)
		input_subset = input[self.norm_idx]
		norm_axis_no0inv = T.switch(theano.tensor.eq(input_subset,0),1,1/input_subset)
		return_mat = theano.tensor.dot(norm_axis_no0inv.reshape((-1,1)), input.reshape((1,-1))).reshape((1,1,self.num_x2, self.num_x1))
		#print(return_mat.shape)
		#getmem()	
		return(return_mat)
	
        def get_output_shape_for(self, input_shape):
		return(1, 1, self.num_x2, self.num_x1)

#Define a custom CNN (called during training). 
def build_custom_cnn(n_out, num_features, normidx, depth, drop_input, drop_hidden=0, width=None, input_var=None, is_image=False, num_responses=1, num_filters=1, filter_size=5, stride_conv=2, drop_out=False):
	if is_image:
		#make_testable_mapping = lapply(fn_make_mat_theano(x,
		#tvar = [fn_make_mat_theano(z, normidx) for z in input_var.eval()]
		#network = lasagne.layers.InputLayer(shape=(None,num_features), input_var=input_var)
		#network = vec_to_image(network, num_x1=num_features, num_x2=num_responses, norm_idx=normidx)
		network = lasagne.layers.InputLayer(shape=(None, 1, num_responses,num_features), input_var=input_var)
	else:
		network = lasagne.layers.InputLayer(shape=(None, 1,num_features), input_var=input_var)
	
	if width is None:
		width = num_features
		print("setting fully connected layer size to number of inputs, {1}".format(num_features))
	
	#The CNN will have a stride of 1, fully connected at pooling. There is usually no dropout in CNNs
	#First a convolution layer, then a max pooling layer
	nonlin = lasagne.nonlinearities.tanh
	conv_nonlin = lasagne.nonlinearities.rectify
	#network = lasagne.layers.MaxPool2DLayer(network, pool_size = (1, 15*stride_conv))
	for _ in range(depth):
		if is_image:
			#If input is 2d image, add a 2D convolution layer, as signal is in 2 dims
			network = lasagne.layers.Conv2DLayer(
				network, num_filters = num_filters, filter_size=(filter_size,filter_size), nonlinearity=conv_nonlin, W=lasagne.init.GlorotUniform(),
				pad = 'full', stride = (stride_conv))
			#Two common variations are F=3, S=2; and more commonly F=2, S=2
			network = lasagne.layers.MaxPool2DLayer(network, pool_size = (stride_conv, stride_conv), stride = stride_conv, pad=0)
		
		else:
			network = lasagne.layers.Conv1DLayer(
				network, num_filters = num_filters, filter_size=filter_size, nonlinearity=conv_nonlin, W=lasagne.init.GlorotUniform(), pad = 'full', stride = stride_conv)
			network = lasagne.layers.MaxPool1DLayer(network, pool_size = 2*stride_conv)

	t2 = lasagne.layers.get_all_layers(network)
	print(map(lambda x: lasagne.layers.get_output_shape(x), t2))
	
	##Then add a fully connected layer, dropout as per input options
	network = lasagne.layers.DenseLayer(network, num_units = num_responses + num_features, nonlinearity = nonlin)
	if drop_hidden:
		network = lasagne.layers.dropout(network, p=drop_hidden)
	
	#Final output layer is softmax, for multiclass output
	softmax = lasagne.nonlinearities.softmax
	network = lasagne.layers.DenseLayer(network, n_out, nonlinearity=softmax)
	if drop_out:
		network = lasagne.layers.dropout(network, p=drop_hidden)
	
	print("---full network---")
	t2 = lasagne.layers.get_all_layers(network)
        print(map(lambda x: lasagne.layers.get_output_shape(x), t2))
	return network

#Define iterator over minibatches of inputs (x) and corresponding labels (y)
def iterate_minibatches(inputs, targets, batchsize, shuffle=False):
	assert len(inputs) == len(targets)
	if shuffle:
		indices = np.arange(len(inputs))
		np.random.shuffle(indices)
	for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
		if shuffle:
			excerpt = indices[start_idx:start_idx + batchsize]
        	else:
			excerpt = slice(start_idx, start_idx + batchsize)
	yield inputs[excerpt], targets[excerpt]

#Define function to reshape 1 dimensional input to an image with 1 channel and width 1
def make_1d_image(x):
	return(x.reshape((x.shape[0],1,x.shape[1],1)))

import theano.sandbox.cuda.basic_ops as sbcuda
import numpy as np
import theano.tensor as T
T.config.floatX = 'float32'
import psutil
import os

def getmem():
	GPUFreeMemoryInBytes = sbcuda.cuda_ndarray.cuda_ndarray.mem_info()[0]
	freeGPUMemInGBs = GPUFreeMemoryInBytes/1024./1024/1024
	print "Your GPU has %s GBs of free memory" % str(freeGPUMemInGBs)
	#An operation is to be executed below
	#testData = shared(np.random.random((5000, 256 * 256)).astype(T.config.floatX), borrow = True)
	#print "The tasks above used %s GBs of your GPU memory. The available memory is %s GBs" % (str(freeGPUMemInGBs - GPUFreeMemoryInBytes/1024./1024/1024), str(GPUFreeMemoryInBytes/1024./1024/1024))
	print("...CPU percent used {0}, physical mem available {1} GB, swap avail {2} GB".format(psutil.cpu_percent(), psutil.virtual_memory().available >> 30, psutil.swap_memory().free >> 30))
	py = psutil.Process(os.getpid())
	memoryUse = py.memory_info()[0]/2.**30
	print("....Process CPU memory use {0} GB".format(memoryUse))

#Define a function to convert a 1d vector into a 2d matrix normalized by an input axis
def normed_idx_vector(idx_norm, data_vector, exclude_normidx=1):
	norm_axis = np.take(data_vector, idx_norm)
	#Since using norm_axis values as scaling factor (dividing), cannot have any zeroes in it
	#Set 0 values in norm_axis vector to 1
	norm_axis[norm_axis==0] = 1
	if exclude_normidx:
		data_vector = np.delete(data_vector, idx_norm)
	##getmem()
	return_dat = np.repeat(data_vector, len(norm_axis)).reshape(len(data_vector), len(norm_axis)).transpose()
	return_dat = return_dat / norm_axis[:, np.newaxis]
	#Shape of returned object is (norming_axis, data_vector)
	#If exclude_normidx is False (or 0), returned object shape is (norming_axis, data_vector without norming elements)
	return(return_dat)

#fn_make_tensor_theano = theano.function(inputs=[xmat, xidx, keep_idx], outputs=[xidx,return_mat])
#tfat = theano.scan(fn=lambda prev_count, next_mat: fn_make_tensor_theano(xmat, prev_step, keep_idx), outputs_info=0, n_steps=xmaxn)

#tfat = theano.scan(fn=fn_make_tensor_theano(xmat, prev_step, keep_idx),outputs_info=[dict(initial=xmat, taps=-1)], sequences=xmat, n_steps=xmaxn) # n_steps=xmaxn)
#xmaxn = T.vector('max_samples', dtype='int8')
#results, updates = theano.scan(fn=fn_make_tensor_theano(fn=lambda prev_step, nmat = fn_make_tensor_theano(xmat, prev_step, keep_idx), outputs_info=[{'initial':xmat}], n_steps=xmaxn)

#tfat = theano.function(inputs=[xmaxn], outputs=xmaxr)

#A = T.matrix("A")
#idx_range = T.arange(A.shape[0])
#updated_img = T.switch(theano.tensor.eq(A, 0), 0, fn_make_mat_theano(A, keep_idx))
#result, updates = theano.scan(fn=lambda idx: fn_make_mat_theano(T.as_tensor_variable(A[idx]), keep_idx), sequences=idx_range)
#fn_mat_img = theano.function(inputs = [A, keep_idx], outputs=[results,updates])
#xs = T.matrix('inputs')

#snum = xs.shape[0]




""" 
cancerscope provides cancer-type predictions for RNA-Seq samples  
"""
import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "..", ".."))
### Global opts
from .config import REF_LABELCOUNTS,SCOPEMODELS_DATADIR, SCOPEMODELS_LIST, SCOPEMODELS_FILELIST_DIR, REF_DISEASECODES, getmodelsdict
from .utils import *
#from cancerscope.utils import *

### Data sources
#from cancerscope.get_models import getmodel, findmodel, downloadmodel
#from cancerscope.get_models import *
from .get_models import getmodel, findmodel, downloadmodel
#from .get_models import *

### Dependent data
#from cancerscope.scope_plots import *
#from cancerscope.scope_ensemble import scopemodel, scope
from .scope_plots import *
from .scope_ensemble import scopemodel, scope

if sys.version_info < (2, 7):
	raise ImportError("Cancerscope module requires Python 2.7*")
__version__ = '0.42'





#Embedded file name: split_input_into_testtrainvalid.py
import datetime, gzip
import csv, math
import numpy as np
import argparse
import re
import collections
import pandas
from argtypes import proper_file, makeintifpossible, int_or_float, check_norm_values
from normalization_functions import norm_minmax, norm_scale, norm_unitscale, norm_scaler_reapply, apply_norm_func
#This was from the older sklearn veresion. Update to version 0.18
#from sklearn.cross_validation import train_test_split, StratifiedShuffleSplit
from sklearn.model_selection import train_test_split, StratifiedShuffleSplit 

import os, json, sys
from six.moves import cPickle
import pickle, csv

def readinputs(datfile, trimlabs, header_status, drugbinderset, sep = '\t', deeplearn=False, feat_startpos=1, sname_startpos=1):
    """This function takes in a matrix of format genes(columns) x samples(rows) and generates an array of labels (Y) and an array of corresponding input vectors (X). The first column is the sample names/labels """
    #If deeplearn, the first column is the samples, the next 4 are the ordered labels
    if deeplearn:
        feat_startpos = 5 ; sname_startpos = 0 # DELETE THIS AFTER DEEPLEARN 
        label_startpos=1 	
    else:
	label_startpos=0
        
    with open(datfile, 'r') as infile:
        reader = csv.reader(infile, delimiter=sep)
        X = []
        Y = []
        labels = []
        header = header_status
	featnames = []
        if drugbinderset:
            ligands = []
            for row in reader:
                if header:
                    header = False
		    featnames = [elem for elem in row[0:]]
                    continue
                row[2:] = [ (float('nan') if elem == 'NA' else float(elem)) for elem in row[2:] ]
                X.append(row[2:])
                ligands.append(row[0])
                Y.append(row[1])
                labels.append(row[1])
            ligands = np.asarray(ligands)
            X = np.asarray(X)
            Y = np.asarray(Y)
            Y = Y.reshape(Y.shape[0], 1)
            if trimlabs:
                labels = [ re.sub('[0-9]*', '', x) for x in labels ]
            returnset = [X,
             Y,
             labels,
             ligands]
        else:
            for row in reader:
                if header:
		    featnames = [elem for elem in row[0:]]
                    header = False
                    continue
                row[feat_startpos:] = [ (float('nan') if elem == 'NA' else float(elem)) for elem in row[feat_startpos:] ]
                X.append(row[feat_startpos:])
                labels.append(row[label_startpos:feat_startpos])
            X = np.asarray(X)
            if (trimlabs) & (deeplearn is False):
		labels = [ re.sub('[0-9]*', '', x[0]) for x in labels ]
	    elif deeplearn is True:
		labels = [[re.sub('[0-9]*', '', i) for i in x] for x in labels ]
            returnset = [X,labels,featnames]
        return returnset

def getnumlabels(labellist):
    """This function takes in an ordered list of format [label1, label1, label2, label3, label3, ...] and generates a numeric label list """
    templist = dict([ (x, labellist.count(x)) for x in collections.OrderedDict.fromkeys(labellist) ])
    clist = [] #templist.values()
    retlist = []
    clabels = []
    retlist = labellist
    index_of_labelcount = 0; 
    #label_keys = collections.OrderedDict.fromkeys(labellist).keys(); 
    #Want an ordered labelling, with max class label = 0, next highest = 1, ....smallest has highest label num
    label_keys = sorted(templist, key=templist.get, reverse=True)
    for i in label_keys:
        if math.isnan(makeintifpossible(i)):
	    labname = index_of_labelcount
        else:
            labname = makeintifpossible(i)
        retlist = [ (labname if ylab == i else ylab) for ylab in retlist ]
        clabels += [labname]
	clist += [templist[i]]
        index_of_labelcount = index_of_labelcount + 1
    retlist = pandas.Series(data=retlist)
    return [templist,
     retlist,
     dict(zip(clabels, clist)),
     dict(zip(label_keys, clabels))]

def splitsets(labelnum, X, testsize, countlist, makevalid, whatsmall, verbose = True, validsize = None, keepsmall=False, deeplearn=False):
    if validsize is None:
        validsize = testsize/2
    print '..Splitting input into training and test, testsize {0} and validation {1}'.format(testsize, validsize)
    trainmat, testmat, trainidx, testidx = splittwogroups(labelnum, X, second_set=testsize, countdict=countlist, whatsmall=whatsmall, verbose=True, keepsmall=keepsmall, deeplearn=deeplearn)
    if makevalid == 1:
        print '...Now splitting training into train and validation'
        trainmat, validmat, trainidx, valididx = splittwogroups(pandas.Series(trainmat['y'][0]), trainmat['X'][0], second_set=validsize, countdict=countlist, whatsmall=whatsmall, verbose=True, keepsmall=keepsmall, deeplearn=deeplearn)
    else:
        testmat, validmat, temp1, valididx = splittwogroups(pandas.Series(testmat['y'][0]), testmat['X'][0], second_set=0, countdict=countlist, whatsmall=whatsmall, verbose=False, keepsmall=keepsmall, deeplearn=deeplearn)
    return [trainmat,validmat,testmat,trainidx,valididx,testidx]

def splittwogroups(inputlabels, inputvectors, second_set, countdict, whatsmall, verbose = True, keepsmall=False,deeplearn=False):
    """This function takes a list of labels and splits them into training and test (or training and validation, whatever). """
    #Inputs are panda Series objects
    if verbose:
        print '  ...Filter threshold is more than {0} samples per class'.format(2 * int(1 / whatsmall))
    inputlabels = pandas.Series(inputlabels)
    classes_high = [ i for i in countdict if countdict[i] > 2 * int(1 / whatsmall) ]
    classes_low = [ i for i in countdict if countdict[i] < 2 * int(1 / whatsmall) + 1 ]
    if 0 & deeplearn:
        lowclass_set = [x for x in inputlabels if x in classes_low]
        highclass_set = [x for x in inputlabels if x in classes_high]
    else:
        lowclass_set = inputlabels[inputlabels.isin(classes_low)]
        highclass_set = inputlabels[inputlabels.isin(classes_high)]
    train_y, test_y = train_test_split(highclass_set, test_size=second_set, random_state=0)
    if(second_set > 0):
	X_high = inputvectors[list(highclass_set.index),:]; y_high = inputlabels[list(highclass_set.index)]
	while((len(set(test_y)) < len(set(highclass_set))) or (len(set(train_y)) < len(set(highclass_set)))):
    		#Need to ensure that after the re-spliting the percentage is ALWAYS maintained
		#StratifiedSplit and other sklearn CV splitters can randomly maintain imbalance for very small classes
	    #	train_y, test_y = train_test_split(highclass_set, test_size = second_set, random_state=123)
    		sss = StratifiedShuffleSplit(n_splits=1, test_size=second_set, random_state=123)
		train_idx, test_idx = sss.split(X_high, y_high)
		train_y = inputlabels[y_high.index[train_idx[0]]]; 
		test_y = inputlabels[y_high.index[test_idx[0]]]
	
    train_x = inputvectors[list(train_y.index), :]
    test_x = inputvectors[list(test_y.index), :]
    small_trainy, small_testy = lowclass_set, lowclass_set
    if keepsmall:
    	train_y = pandas.concat([train_y, small_trainy])
	train_x = np.concatenate((train_x, inputvectors[list(small_trainy.index), :]))
    	if second_set > 0:
		test_y = pandas.concat([test_y, small_testy])
    		test_x = np.concatenate((test_x, inputvectors[list(small_testy.index), :]))
	
    if verbose:
        print '  ...{0} train large, {1} train small \n  ...{2} test large, {3} test small'.format(len(train_y), len(small_trainy), len(test_y), len(small_testy))

    trainmat = {'X': [np.asarray(train_x)],
     'y': [np.asarray(train_y)]}
    testmat = {'X': [np.asarray(test_x)],
     'y': [np.asarray(test_y)]}
    return [trainmat,testmat,train_y.index,test_y.index]


def takeawaytest(inputlabels, inputvectors, ligandlist, classligands = None):
    if classligands is None:
        classligs = list()
        for i in set(inputlabels):
            print 'Class %s' % i
            print set(ligandlist[np.where(inputlabels == i)])
            if len(set(ligandlist[np.where(inputlabels == i)])) < 2:
                print 'WARNING: This set only has 1 ligand, it will be split into train and valid, but not tested.'
            else:
                selectfortest = np.random.choice(ligandlist[np.where(inputlabels == i)])
                classligs.append(selectfortest)

    else:
        classligs = classligands
    test_idx = np.where(np.in1d(ligandlist, classligs))[0]
    train_idx = np.where(np.invert(np.in1d(ligandlist, classligs)))[0]
    test_y = inputlabels[test_idx]
    test_x = inputvectors[test_idx]
    train_y = inputlabels[train_idx]
    train_x = inputvectors[train_idx]
    trainmat = {'X': [np.asarray(train_x)],
     'y': [np.asarray(train_y)]}
    testmat = {'X': [np.asarray(test_x)],
     'y': [np.asarray(test_y)]}
    return [trainmat,
     testmat,
     train_y.index,
     test_y.index,
     classligs]


def labelweighing(labcountdict, labels_set):
    tlist = list(dict(labcountdict).values())
    maxcount = max(tlist)
    tlist2 = [ float(maxcount) / x for x in tlist ]
    tlist3 = [ math.log(x, 2) for x in tlist2 ]
    tlist = dict(zip(labels_set.values(), tlist3))
    return tlist

class mydatstuff(object):

    def __init__(self, infile, outdir, trimflag, keepheader, drugbinderflag, fulltrainflag, keepsmall, normtype=None, testsize = 0.1, normby = 1, outprefix = None , makevalid = 1, classligands = None,verbose=False, x=None, labs=None,ligs=None, deeplearn=False):
        self.input = infile
        self.outdir = outdir
	self.deeplearn = deeplearn
        if(x is not None):
		self.X, self.labels, self.ligands = x, labs, ligs
	else:
	    self.X, self.labels, self.ligands = readinputs(infile, trimflag, keepheader, drugbinderset=drugbinderflag, deeplearn=deeplearn)
	    
	    t = pandas.DataFrame(self.X); t = t.dropna(how='all', axis=1); self.X = np.asarray(t)
        
	if deeplearn:
		self.labels = np.asarray(self.labels)
		lc=[]; ln=[]; cl=[]; ls=[]; lw=[]
        	for i in range(np.shape(self.labels)[1]):
			catlist = getnumlabels(list(self.labels[:,i])) 
			lc.append(catlist[0]); ln.append(catlist[1]); cl.append(catlist[2]); ls.append(catlist[3])
			print("Total classes level {0} are {1}\n".format(str(i), str(max(catlist[1])+1)))
			lw.append(labelweighing(catlist[0], catlist[3]))
		
		merged_lab = [a + "-" + b + "-" + c + "-" + d for a,b,c,d in zip(self.labels[:,0], self.labels[:,1],self.labels[:,2],self.labels[:,3])]
		self.labelcount_unrolled = lc; self.labelnum_unrolled =ln; self.countlist_unrolled=cl; self.labels_set_unrolled = ls
		flip_labs = [{str(v).zfill(2): k for k, v in my_map.items()} for my_map in self.labels_set_unrolled]
			
		self.labelnum = [str(a).zfill(2) + str(b).zfill(2) + str(c).zfill(2) + str(d).zfill(2) for a,b,c,d in zip(self.labelnum_unrolled[0], self.labelnum_unrolled[1],self.labelnum_unrolled[2],self.labelnum_unrolled[3])]
		
		self.labels_set = [flip_labs[0][str(a).zfill(2)] + "-" + flip_labs[1][str(b).zfill(2)] + "-" + flip_labs[2][str(c).zfill(2)] + "-" + flip_labs[3][str(d).zfill(2)] for a,b,c,d in zip(self.labelnum_unrolled[0], self.labelnum_unrolled[1],self.labelnum_unrolled[2],self.labelnum_unrolled[3])]
		self.labels_set = dict(set(zip(self.labels_set,self.labelnum)))
		self.labelcount = dict(collections.Counter(merged_lab))
		self.countlist = dict(collections.Counter(self.labelnum))	
		self.merged_labels = merged_lab
		print("A sample list of y values, mapped to int levels {0}".format(self.labelnum[1:10]))
		
		self.labels_weighted = labelweighing(self.labelcount, self.labels_set)
		#self.labelcount_unrolled = lc; self.labelnum_unrolled =ln; self.countlist_unrolled=cl; self.labels_set_unrolled = ls
	else:
		self.labelcount, self.labelnum, self.countlist, self.labels_set = getnumlabels(self.labels)
		print 'Total classes are {0}'.format(str(max(self.labelnum) + 1))
        	self.labels_weighted = labelweighing(self.labelcount, self.labels_set)
       
	 
	self.outpref = outprefix
        self.normtype = normtype
        self.normby = int(normby)
        self.testsize = testsize
        self.makevalid = makevalid
	self.keep_small = keepsmall
	self.whatsmall=0.2
	self.verbose=verbose
	if self.normtype == 'none':
            normedby = '_normed' + self.normtype
        elif self.normby == 1:
            normedby = '_normedSamples_' + self.normtype
        else:
            normedby = '_normedFeatures_' + self.normtype
        self.outpref = self.outpref + normedby + '_makevalid' + str(self.makevalid) + '_keepsmall' + str(self.keep_small)
	self.drugbinderflag = drugbinderflag; self.fulltrainflag=fulltrainflag; 
	
    def splitvalidtest(self,returnsets=False):
        if self.drugbinderflag:
            if self.fulltrainflag:
                self.trainmat, self.validmat, self.testmat, self.trainidx, self.valididx, self.testidx = splitsets(labelnum=self.labelnum, X=self.X, testsize=0, countlist=self.countlist, makevalid=self.makevalid, validsize=self.testsize, keepsmall=self.keep_small, whatsmall=self.whatsmall, verbose=self.verbose)
                self.testlabs = [0, 0]
            else:
                self.trainmat, self.testmat, self.trainidx, self.testidx, self.testlabs = takeawaytest(self.labelnum, self.X, self.ligands, classligands)
                print '...Test set formed with ligands: {0}'.format(self.testlabs)
                if self.makevalid == 1:
                    print '...Generating validation set as well'
                    self.trainmat, self.validmat, self.trainidx, self.valididx = splittwogroups(pandas.Series(self.trainmat['y'][0]), self.trainmat['X'][0], second_set=self.testsize, countdict=self.countlist, whatsmall=self.whatsmall, keepsmall=self.keep_small)
                else:
                    self.trainmat, self.validmat, self.temp, self.valididx = splittwogroups(pandas.Series(self.trainmat['y'][0]), self.trainmat['X'][0], second_set=0, countdict=self.countlist, whatsmall=self.whatsmall, verbose=False, keepsmall=self.keep_small)
	else:
		if self.fulltrainflag:
			self.trainmat, self.validmat, self.testmat, self.trainidx, self.valididx, self.testidx = splitsets(labelnum=self.labelnum, X=self.X, testsize=0, countlist=self.countlist, makevalid=self.makevalid, validsize=self.testsize, keepsmall=self.keep_small, whatsmall=self.whatsmall, verbose=self.verbose, deeplearn=self.deeplearn)
            	else:
			self.trainmat, self.validmat, self.testmat, self.trainidx, self.valididx, self.testidx = splitsets(labelnum=self.labelnum, X=self.X, testsize=self.testsize, whatsmall=self.whatsmall, countlist=self.countlist, makevalid=self.makevalid, keepsmall=self.keep_small,verbose=self.verbose, deeplearn=self.deeplearn)
        
        print 'Training set has {0} classes \n .....{1} in test \n .....{2} in validation'.format(len(list(set(self.trainmat['y'][0]))), len(list(set(self.testmat['y'][0]))), len(list(set(self.validmat['y'][0]))))
        if self.fulltrainflag:
            self.testmat = self.validmat
            self.testidx = self.valididx
		
    def normalize(self):
        print("NORMALIZATION: {0}".format(self.normtype))
	if self.normtype == 'none':
            print 'Not normalizing data'
        elif self.normby == 1:
            print '...normalizing each sample. This is done agnostic of the type of dataset (train/test/valid).'
            self.trainmat['X'][0] = np.asarray(apply_norm_func(norm_func=self.normtype, xdat=self.trainmat['X'][0], bysamples=1))
            if (self.testmat['X'][0]).shape[0] > 0:
	        self.testmat['X'][0] = np.asarray(apply_norm_func(norm_func=self.normtype, xdat=self.testmat['X'][0], bysamples=1))
                self.validmat['X'][0] = np.asarray(apply_norm_func(norm_func=self.normtype, xdat=self.validmat['X'][0], bysamples=1))
        else:
            print '...normalizing each feature. This is done within the train/valid/test sets separately.'
            self.trainmat['X'][0], self.testmat['X'][0], self.validmat['X'][0] = apply_norm_func(norm_func=self.normtype, xdat=[self.trainmat['X'][0], self.testmat['X'][0], self.validmat['X'][0]], bysamples=0)

    def return_dataset(self):
        #Note that self.ligands is the set of features (if header provided for input txt), or an empty set, for general purposes.
	#It is the set of ligands when reading the drugbinder set
	if self.deeplearn:
		return([[self.trainmat['X'][0], self.trainmat['y'][0]], [self.validmat['X'][0], self.validmat['y'][0]], [self.testmat['X'][0], self.testmat['y'][0]]],self.labels_set, dict(self.labels_weighted), self.ligands)
	else:
		return([[self.trainmat['X'][0], self.trainmat['y'][0]], [self.validmat['X'][0], self.validmat['y'][0]], [self.testmat['X'][0], self.testmat['y'][0]]],self.labels_set, dict(self.labels_weighted), self.ligands)

    def writeout(self,save_dat=False, write_idx=True):
        if(self.input is None):
                filename = "predefined_input"
	else:
		filename = os.path.splitext(os.path.basename(self.input.rsplit('.', 1)[0]))[0]
        self.outpref = filename + '_' + self.outpref
        self.outdir = os.path.join(os.path.abspath(self.outdir) + '/' + self.outpref + '_' + datetime.datetime.now().strftime('%Y-%m-%d'))
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
        print (self.trainmat['X'][0].shape, self.validmat['X'][0].shape, self.testmat['X'][0].shape)
	print("Writing processed data features to {0}".format(self.outdir))
        if save_dat:
		outall = gzip.open(os.path.join(self.outdir, ''.join([self.outpref, '.pkl.gz'])), 'wb')
		cPickle.dump([[self.trainmat['X'][0], self.trainmat['y'][0]], [self.validmat['X'][0], self.validmat['y'][0]], [self.testmat['X'][0], self.testmat['y'][0]]], outall, protocol=cPickle.HIGHEST_PROTOCOL)
		outall.close()
	
	if write_idx:
		with open(os.path.join(self.outdir, 'trainingidxs.txt'), 'w') as f:
			[ f.write('{0}\n'.format(x)) for x in self.trainidx ]
		with open(os.path.join(self.outdir, 'testidxs.txt'), 'w') as f:
			[ f.write('{0}\n'.format(x)) for x in self.testidx ]
		with open(os.path.join(self.outdir, 'valididxs.txt'), 'w') as f:
			[ f.write('{0}\n'.format(x)) for x in self.valididx ]
        
	with open(os.path.join(self.outdir, ''.join(['dict_labels_', self.outpref, '.txt'])), 'w') as f:
            [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labels_set.items() ]
        with open(os.path.join(self.outdir, ''.join(['dict_labcount_', self.outpref, '.txt'])), 'w') as f:
            [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in dict(self.labelcount).items() ]
        with open(os.path.join(self.outdir, ''.join(['dict_labweights_', self.outpref, '.txt'])), 'w') as f:
            [ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labels_weighted.items() ]

	if self.deeplearn:
		with open(os.path.join(self.outdir, ''.join(['dict_labels_', self.outpref, 'unrolled.txt'])), 'w') as f:
			[ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labels_set_unrolled.items() ]
		with open(os.path.join(self.outdir, ''.join(['dict_labcount_', self.outpref, 'unrolled.txt'])), 'w') as f:
			[ f.write('{0}\t{1}\n'.format(key, value)) for key, value in dict(self.labelcount_unrolled).items() ]
		with open(os.path.join(self.outdir, ''.join(['dict_labweights_', self.outpref, 'unrolled.txt'])), 'w') as f:
			[ f.write('{0}\t{1}\n'.format(key, value)) for key, value in self.labels_weighted_unrolled.items() ]

if 0:
    myargs = getargs()
    logfile = open(myargs.outdir[0] + '/logfile.txt', 'a')
    datfile = myargs.input[0]
    outpref = myargs.pref[0]
    makevalid = myargs.make_valid
    testsize = int_or_float(myargs.testsize[0])
    outdir = myargs.outdir[0]
    print 'Reading input from {0} \n Outprefix is {1} \n Creating validation set? {2} \n Size of test set (and validation set, if any) {3} \n'.format(datfile, outpref, makevalid, testsize)
    if myargs.normtype is None:
        normtopass = 'none'
        norm_sample = 0
    else:
        normtopass = myargs.normtype[0]
        norm_sample = myargs.normalize[0]
    
    if myargs.ligands is None:
        dataseparator = mydatstuff(infile=myargs.input[0], outdir=myargs.outdir[0], trimflag=myargs.trimlabels, keepheader=myargs.header_row, testsize=int_or_float(myargs.testsize[0]), normby=norm_sample, outprefix=myargs.pref[0], normtype=normtopass, makevalid=myargs.make_valid, drugbinderflag=myargs.drugbinder, fulltrainflag=myargs.fulltrainset, keepsmall=myargs.keep_small)
    else:
        print '...Using user defined classes as tests, {0}'.format(myargs.ligands)
        dataseparator = mydatstuff(infile=myargs.input[0], outdir=myargs.outdir[0], trimflag=myargs.trimlabels, keepheader=myargs.header_row, testsize=int_or_float(myargs.testsize[0]), normby=norm_sample, outprefix=myargs.pref[0], normtype=normtopass, makevalid=myargs.make_valid, drugbinderflag=myargs.drugbinder, classligands=myargs.ligands, fulltrainflag=myargs.fulltrainset, keepsmall=myargs.keep_small)
    dataseparator.normalize()
    dataseparator.writeout()
